# (C) Copyright 2025 by Rocky Bernstein
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
RustPython 3.13 bytecode opcodes for version 0.40. There are other Rust 3.13 with different opcodes!
"""

#FIXME: this needs a lot of going over.

from typing import Dict, List, Optional, Tuple

import xdis.opcodes.opcode_3x.opcode_311 as opcode_311

# import xdis.opcodes.opcode_313 as opcode_313
from xdis.opcodes.base import (
    VARYING_STACK_INT,
    binary_op,
    call_op,
    compare_op,
    const_op,
    def_op,
    finalize_opcodes,
    free_op,
    init_opdata,
    jabs_op,
    jrel_op,
    local_op,
    name_op,
    nargs_op,
    store_op,
    unary_op,
    update_pj3,
    varargs_op,
)
from xdis.opcodes.format.extended import extended_format_binary_op
from xdis.opcodes.opcode_3x.opcode_313 import opcode_arg_fmt313, opcode_extended_fmt313
from xdis.opcodes.opcode_rust.base import cmp_op as cmp_op_rust
from xdis.version_info import PythonImplementation

python_implementation = PythonImplementation("RustPython")
version_tuple = (3, 13)

# oppush[op] => number of stack entries pushed
oppush: List[int] = [0] * 256

# oppop[op] => number of stack entries popped
oppop: List[int] = [0] * 256

# opmap[opcode_name] => opcode_number
opmap: Dict[str, int] = {}

## pseudo opcodes (used in the compiler) mapped to the values
## they can become in the actual code.
_pseudo_ops = {}

_nb_ops = [
    ("NB_ADD", "+"),
    ("NB_AND", "&"),
    ("NB_FLOOR_DIVIDE", "//"),
    ("NB_LSHIFT", "<<"),
    ("NB_MATRIX_MULTIPLY", "@"),
    ("NB_MULTIPLY", "*"),
    ("NB_REMAINDER", "%"),
    ("NB_OR", "|"),
    ("NB_POWER", "**"),
    ("NB_RSHIFT", ">>"),
    ("NB_SUBTRACT", "-"),
    ("NB_TRUE_DIVIDE", "/"),
    ("NB_XOR", "^"),
    ("NB_INPLACE_ADD", "+="),
    ("NB_INPLACE_AND", "&="),
    ("NB_INPLACE_FLOOR_DIVIDE", "//="),
    ("NB_INPLACE_LSHIFT", "<<="),
    ("NB_INPLACE_MATRIX_MULTIPLY", "@="),
    ("NB_INPLACE_MULTIPLY", "*="),
    ("NB_INPLACE_REMAINDER", "%="),
    ("NB_INPLACE_OR", "|="),
    ("NB_INPLACE_POWER", "**="),
    ("NB_INPLACE_RSHIFT", ">>="),
    ("NB_INPLACE_SUBTRACT", "-="),
    ("NB_INPLACE_TRUE_DIVIDE", "/="),
    ("NB_INPLACE_XOR", "^="),
]

hasexc = []

loc = locals()

init_opdata(loc, from_mod=None, version_tuple=version_tuple)
# Change cmp_op set in init_opdata to be Rust's version of this.
cmp_op = cmp_op_rust

loc["opname"].extend([f"<{i}>" for i in range(256, 267)])
loc["oppop"].extend([0] * 11)
loc["oppush"].extend([0] * 11)

hasarg = []
hasconst = []
hasname = []
hasjrel = []
hasjabs = []
haslocal = []
hascompare = []
hasfree = []
hasexc = []

oplists = [
    loc["hasarg"],
    loc["hasconst"],
    loc["hasname"],
    loc["hasjrel"],
    loc["hasjabs"],
    loc["haslocal"],
    loc["hascompare"],
    loc["hasfree"],
    loc["hasexc"],
]

# add new table "hasjump"
loc.update({"hasjump": []})
loc["hasjrel"] = loc["hasjump"]

def pseudo_op(name: str, op: int, real_ops: list):
    def_op(loc, name, op)
    _pseudo_ops[name] = real_ops
    # add the pseudo opcode to the lists its targets are in
    for oplist in oplists:
        res = [opmap[rop] in oplist for rop in real_ops]
        if any(res):
            # FIXME: for some reason JUMP_FORWARD appears to be
            # listed as a free op. It isn't.
            # if not all(res):
            #     breakpoint()
            # assert all(res)
            oplist.append(op)


# See RustPython/compiler/core/src/bytecode.rs Instruction
# fmt: off

#            OP NAME                   OPCODE   POP PUSH
#----------------------------------------------------------
def_op(loc, "BEFORE_ASYNC_WITH",            0)
binary_op(loc, "BINARY_OP",                 1,  1, 1)
binary_op(loc, "BINARY_SUBSCR",             2)
jabs_op(loc,   "BREAK_LOOP",                3,  0, 0)
varargs_op(loc, "BUILD_LIST_FROM_TUPLES",   4, 1, VARYING_STACK_INT)       # Number of list items
varargs_op(loc, "BUILD_LIST",               5, 1, VARYING_STACK_INT)       # Number of list items
varargs_op(loc, "BUILD_MAP_FOR_CALL",       6, 1, VARYING_STACK_INT)       # Number of dict entries
varargs_op(loc, "BUILD_MAP",                7, 1, VARYING_STACK_INT)       # Number of dict entries
varargs_op(loc, "BUILD_SET_FROM_TUPLES",    8, 1, VARYING_STACK_INT)       # Number of list items
varargs_op(loc, "BUILD_SET",                9, 1, VARYING_STACK_INT)       # Number of set items
def_op(loc, "BUILD_SLICE",                 10)
varargs_op(loc, "BUILD_STRING",            11, VARYING_STACK_INT)      # Number of tuple items
varargs_op(loc, "BUILD_TUPLE_FROM_ITER",   12, 1, VARYING_STACK_INT)      # Number of tuple items
varargs_op(loc, "BUILD_TUPLE_FROM_TUPLES", 13, 1, VARYING_STACK_INT)      # Number of tuple items
varargs_op(loc, "BUILD_TUPLE",             14, 1, VARYING_STACK_INT)      # Number of tuple items

call_op(loc, "CALL_FUNCTION_EX",           15)  # Flags
call_op(loc, "CALL_FUNCTION_KW",           16)
call_op(loc, "CALL_FUNCTION",              17)
call_op(loc, "CallIntrinsic1",             18)
call_op(loc, "CallIntrinsic2",             19)
call_op(loc, "CALL_METHOD_EX",             20)
call_op(loc, "CALL_METHOD_KW",             21)
call_op(loc, "CALL_METHOD",                22)
jrel_op(loc, "CHECK_EG_MATCH",             23,   0, 0)
compare_op(loc, "COMPARE_OP",              24,  2, 1)  # Comparison operator
compare_op(loc, "CONTAINS_OP",             25,  2, 1)
jabs_op(loc, "CONTINUE_LOOP",              26,  2, 1)
def_op(loc,    "ConvertValue",             27,  0, 0)
def_op(loc,    "COPY",                     28,  0, 0)

name_op(loc,   "DELETE_ATTR",              29)        # Local variable number is in operand
free_op(loc,   "DELETE_DEREF",             30,  0, 0)
local_op(loc,  "DELETE_FAST",              31,  0, 0) # Local variable number is in operand
name_op(loc,   "DELETE_GLOBAL",            32)         # ""
local_op(loc,  "DELETE_LOCAL",             33,  0, 0)
def_op(loc,    "DELETE_SUBSCR",            34,  0, 0)
varargs_op(loc, "DICT_UPDATE",             35)        # Number of dict entries

def_op(loc, "END_ASYNC_FOR",               36)
def_op(loc, "END_FINALLY",                 37)
def_op(loc, "EnterFinally",                38)
def_op(loc, "EXTENDED_ARG",                39)

jabs_op(loc, "FOR_ITER",                   40)
def_op(loc, "FormatSimple",                41, VARYING_STACK_INT, VARYING_STACK_INT)
def_op(loc, "FormatWithSpec",              42, VARYING_STACK_INT, VARYING_STACK_INT)
def_op(loc, "GET_AITER",                   43)
def_op(loc, "GET_ANEXT",                   44)
def_op(loc, "GET_AWAITABLE",               45)
def_op(loc, "GET_ITER",                    46)
def_op(loc, "GetLen",                      47)


name_op(loc,   "IMPORT_FROM",              48)
name_op(loc,   "IMPORT_NAME",              49)
def_op(loc,    "IsOp",                     50,  0, 1)

jabs_op(loc, "JUMP_IF_FALSE_OR_POP",       51) # "
jabs_op(loc, "JumpIfNotExcMatch",          52) # Number of words to skip
jabs_op(loc, "JUMP_IF_TRUE_OR_POP",        53) # Number of words to skip
jabs_op(loc,  "JUMP",                      54) # Number of words to skip

def_op(loc,   "LIST_APPEND",               55)
name_op(loc,  "LOAD_ATTR",                 56)       # Index in name list
def_op(loc,   "LOAD_BUILD_CLASS",          57)
def_op(loc,   "LOAD_CLASS_DEREF",          59)
free_op(loc,  "LOAD_CLOSURE",              59)
const_op(loc, "LOAD_CONST",                60,  0, 1)
free_op(loc,  "LOAD_DEREF",                61)
local_op(loc, "LOAD_FAST",                 62)
name_op(loc,  "LOAD_GLOBAL",               63)
def_op(loc,   "LOAD_METHOD",               64)
name_op(loc,  "LOAD_NAME",                 65)

nargs_op(loc, "MAKE_FUNCTION",             66, VARYING_STACK_INT)    # Flags
def_op(loc, "MAP_ADD",                     67)
def_op(loc, "MATCH_CLASS",                 68)
def_op(loc, "MATCH_KEYS",                  69)
def_op(loc, "MATCH_MAPPING",               70)
def_op(loc, "MATCH_SEQUENCE",              71)

def_op(loc, "NOP",                         72,  0, 0)

def_op(loc, "POP_BLOCK",                   73)
def_op(loc, "POP_EXCEPT",                  74)
jabs_op(loc, "POP_JUMP_IF_FALSE",          75)
jabs_op(loc, "POP_JUMP_IF_TRUE",           76)
def_op(loc,  "POP_TOP",                    77,  1, 0)

def_op(loc, "RAISE",                       78)
def_op(loc, "RESUME",                      79,   0, 0)
const_op(loc, "RETURN_CONST",              80)
def_op(loc, "RETURN_VALUE",                81)
def_op(loc, "REVERSE",                     82)

def_op(loc, "SET_ADD",                     83)
def_op(loc, "SET_FUNCTION_ATTR",           84)
def_op(loc, "SETUP_ANNOTATION",            85)
def_op(loc, "SETUP_ASYNC_WITH",            86, 0, 1)
def_op(loc, "SETUP_EXCEPT",                87)
jabs_op(loc, "SETUP_FINALLY",              88)
jabs_op(loc, "SETUP_LOOP",                 89)
jabs_op(loc, "SETUP_WITHP",                90)
name_op(loc,   "STORE_ATTR",               91)        # Index in name list
store_op(loc,  "STORE_DEREF",              92,  1, 0, is_type="free")
store_op(loc,  "STORE_FAST",               93,  1, 0, is_type="local")  # Local variable
name_op(loc,   "STORE_GLOBAL",             94)                           # ""
store_op(loc,  "STORE_LOCAL",              95,  4,  0)
def_op(loc,    "STORE_SUBSCR",             96,  0, 0)
binary_op(loc,  "SUBCRIPT",                97)
def_op(loc,     "SWAP",                    98,  1, 1)

compare_op(loc, "TEST_OP",                 99,  2, 1)  # test operator
def_op(loc,     "TO_BOOL",                100,  1, 1)

unary_op(loc,  "UNARY_OP",                101,  1, 1)
def_op(loc, "UNPACK_EX",                  102, VARYING_STACK_INT, VARYING_STACK_INT)
varargs_op(loc, "UNPACK_SEQUENCE",        103, 1, VARYING_STACK_INT)   # Number of tuple items

def_op(loc, "WithCleanupFinish",          104)
def_op(loc, "WithCleanupStart",           105)

def_op(loc, "YIELD",                      106)
def_op(loc, "YIELD_FROM",                 107)


EXTENDED_ARG = 103

# fmt: on

_specializations = {
    "BINARY_OP": [
        "BINARY_OP_ADAPTIVE",
        "BINARY_OP_ADD_FLOAT",
        "BINARY_OP_ADD_INT",
        "BINARY_OP_ADD_UNICODE",
        "BINARY_OP_INPLACE_ADD_UNICODE",
        "BINARY_OP_MULTIPLY_FLOAT",
        "BINARY_OP_MULTIPLY_INT",
        "BINARY_OP_SUBTRACT_FLOAT",
        "BINARY_OP_SUBTRACT_INT",
    ],
    "BINARY_SUBSCR": [
        "BINARY_SUBSCR_ADAPTIVE",
        "BINARY_SUBSCR_DICT",
        "BINARY_SUBSCR_GETITEM",
        "BINARY_SUBSCR_LIST_INT",
        "BINARY_SUBSCR_TUPLE_INT",
    ],
    "CALL": [
        "CALL_ADAPTIVE",
        "CALL_PY_EXACT_ARGS",
        "CALL_PY_WITH_DEFAULTS",
        "CALL_BOUND_METHOD_EXACT_ARGS",
        "CALL_BUILTIN_CLASS",
        "CALL_BUILTIN_FAST_WITH_KEYWORDS",
        "CALL_METHOD_DESCRIPTOR_FAST_WITH_KEYWORDS",
        "CALL_NO_KW_BUILTIN_FAST",
        "CALL_NO_KW_BUILTIN_O",
        "CALL_NO_KW_ISINSTANCE",
        "CALL_NO_KW_LEN",
        "CALL_NO_KW_LIST_APPEND",
        "CALL_NO_KW_METHOD_DESCRIPTOR_FAST",
        "CALL_NO_KW_METHOD_DESCRIPTOR_NOARGS",
        "CALL_NO_KW_METHOD_DESCRIPTOR_O",
        "CALL_NO_KW_STR_1",
        "CALL_NO_KW_TUPLE_1",
        "CALL_NO_KW_TYPE_1",
    ],
    "COMPARE_OP": [
        "COMPARE_OP_ADAPTIVE",
        "COMPARE_OP_FLOAT_JUMP",
        "COMPARE_OP_INT_JUMP",
        "COMPARE_OP_STR_JUMP",
    ],
    "EXTENDED_ARG": [
        "EXTENDED_ARG_QUICK",
    ],
    "FOR_ITER": [
        "FOR_ITER_ADAPTIVE",
        "FOR_ITER_LIST",
        "FOR_ITER_RANGE",
    ],
    "JUMP_BACKWARD": [
        "JUMP_BACKWARD_QUICK",
    ],
    "LOAD_ATTR": [
        "LOAD_ATTR_ADAPTIVE",
        # These potentially push [NULL, bound method] onto the stack.
        "LOAD_ATTR_CLASS",
        "LOAD_ATTR_GETATTRIBUTE_OVERRIDDEN",
        "LOAD_ATTR_INSTANCE_VALUE",
        "LOAD_ATTR_MODULE",
        "LOAD_ATTR_PROPERTY",
        "LOAD_ATTR_SLOT",
        "LOAD_ATTR_WITH_HINT",
        # These will always push [unbound method, self] onto the stack.
        "LOAD_ATTR_METHOD_LAZY_DICT",
        "LOAD_ATTR_METHOD_NO_DICT",
        "LOAD_ATTR_METHOD_WITH_DICT",
        "LOAD_ATTR_METHOD_WITH_VALUES",
    ],
    "LOAD_CONST": [
        "LOAD_CONST__LOAD_FAST",
    ],
    "LOAD_FAST": [
        "LOAD_FAST__LOAD_CONST",
        "LOAD_FAST__LOAD_FAST",
    ],
    "LOAD_GLOBAL": [
        "LOAD_GLOBAL_ADAPTIVE",
        "LOAD_GLOBAL_BUILTIN",
        "LOAD_GLOBAL_MODULE",
    ],
    "RESUME": [
        "RESUME_QUICK",
    ],
    "STORE_ATTR": [
        "STORE_ATTR_ADAPTIVE",
        "STORE_ATTR_INSTANCE_VALUE",
        "STORE_ATTR_SLOT",
        "STORE_ATTR_WITH_HINT",
    ],
    "STORE_FAST": [
        "STORE_FAST__LOAD_FAST",
        "STORE_FAST__STORE_FAST",
    ],
    "STORE_SUBSCR": [
        "STORE_SUBSCR_ADAPTIVE",
        "STORE_SUBSCR_DICT",
        "STORE_SUBSCR_LIST_INT",
    ],
    "UNPACK_SEQUENCE": [
        "UNPACK_SEQUENCE_ADAPTIVE",
        "UNPACK_SEQUENCE_LIST",
        "UNPACK_SEQUENCE_TUPLE",
        "UNPACK_SEQUENCE_TWO_TUPLE",
    ],
}
_specialized_instructions = [
    opcode for family in _specializations.values() for opcode in family
]
_specialization_stats = [
    "success",
    "failure",
    "hit",
    "deferred",
    "miss",
    "deopt",
]

_cache_format = {
    "LOAD_GLOBAL": {
        "counter": 1,
        "index": 1,
        "module_keys_version": 2,
        "builtin_keys_version": 1,
    },
    "BINARY_OP": {
        "counter": 1,
    },
    "UNPACK_SEQUENCE": {
        "counter": 1,
    },
    "COMPARE_OP": {
        "counter": 1,
        "mask": 1,
    },
    "BINARY_SUBSCR": {
        "counter": 1,
        "type_version": 2,
        "func_version": 1,
    },
    "FOR_ITER": {
        "counter": 1,
    },
    "LOAD_ATTR": {
        "counter": 1,
        "version": 2,
        "keys_version": 2,
        "descr": 4,
    },
    "STORE_ATTR": {
        "counter": 1,
        "version": 2,
        "index": 1,
    },
    "CALL": {
        "counter": 1,
        "func_version": 2,
        "min_args": 1,
    },
    "STORE_SUBSCR": {
        "counter": 1,
    },
}

_inline_cache_entries = [
    sum(_cache_format.get(opname[opcode], {}).values()) for opcode in range(256)
]

findlinestarts = opcode_311.findlinestarts

def extended_format_BINARY_OP(opc, instructions) -> Tuple[str, Optional[int]]:
    opname = _nb_ops[instructions[0].argval][1]
    if opname == "%":
        opname = "%%"
    elif opname == "%=":
        opname = "%%="
    return extended_format_binary_op(opc, instructions, f"%s {opname} %s")


opcode_extended_fmt313rust = {}
opcode_arg_fmt = opcode_arg_fmt13rust = {}

### update arg formatting
opcode_extended_fmt = opcode_extended_fmt312rust = {
    **opcode_extended_fmt313,
    **{
        "BINARY_OP": extended_format_BINARY_OP,
    },
}


update_pj3(globals(), loc, is_rust=True)
finalize_opcodes(loc)



"""
Copyright (c) 2020-2025 by Rocky Bernstein

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

import platform
import sys
from enum import Enum
from typing import Tuple

PYTHON3 = sys.version_info >= (3, 0)

PYTHON_VERSION_TRIPLE = tuple(sys.version_info[:3])
PYTHON_VERSION_STR = "%s.%s" % (sys.version_info[0], sys.version_info[1])


class PythonImplementation(Enum):
    """
    Enumeration of Python interpreter implementations. Each member's value is the
    canonical string returned by platform.python_implementation() for that implementation.
    """

    CPython = "CPython"
    PyPy = "PyPy"
    Graal = "Graal"
    RustPython = "RustPython"
    Jython = "Jython"
    Other = "Other"

    def __str__(self) -> str:
        """
        Return the string value of the implementation. This makes str(PythonImplemtation.*)
        return the underlying implementation string (e.g. "CPython", "Graal", ...).
        """
        return self.value


def get_python_implementation(
    implementation: str = platform.python_implementation(),
) -> PythonImplementation:
    """
    Detect the current Python implementation and return the corresponding
    PlatformImplemtation enum member.
    """
    # Match common exact names first
    if implementation == "CPython":
        return PythonImplementation.CPython
    elif implementation == "PyPy":
        return PythonImplementation.PyPy
    elif implementation == "RustPython":
        return PythonImplementation.RustPython
    elif implementation == "Jython":
        return PythonImplementation.Jython
    # Graal may appear as "GraalVM" or include "Graal" in some environments
    elif "Graal" in implementation:
        return PythonImplementation.Graal
    # If nothing matched, return Other.
    return PythonImplementation.Other


PYTHON_IMPLEMENTATION = get_python_implementation()
IS_C_PYTHON = PYTHON_IMPLEMENTATION == PythonImplementation.CPython
IS_GRAAL = PYTHON_IMPLEMENTATION == PythonImplementation.Graal
IS_PYPY = PYTHON_IMPLEMENTATION == PythonImplementation.PyPy
IS_RUST = PYTHON_IMPLEMENTATION == PythonImplementation.RustPython


def version_tuple_to_str(
    version_tuple: Tuple[int, ...] = PYTHON_VERSION_TRIPLE,
    start: int = 0,
    end: int = 3,
    delimiter: str = ".",
) -> str:
    """
    Turn a version tuple, e.g. (3,2,6), into a dotted string, e.g. "3.2.6".

    ``version_tuple`` is a tuple similar to what is might be returned in
    tuple(sys.version_info[:3]), however, the parts in the tuple could anything that
    has a str() method. By default, and often the length is 3, but in practice
    it could be other lengths.

    ``end`` is the length of version_tuple that you want to use.
    delimiter is what string to put in the between components.
    """
    return delimiter.join([str(v) for v in version_tuple[start:end]])


def version_str_to_tuple(python_version: str, length: int = 2) -> tuple:
    return tuple([int(v) for v in python_version.split(".")[:length]])



# (C) Copyright 2025 by Rocky Bernstein
"""PYPY 3.11 opcodes

PyPy's 3.11 opcodes are more like CPython's 3.10's opcode.py than CPython 3.11 opcodes.

We add some classification of stack usage and information for
formatting instructions.

"""

import xdis.opcodes.opcode_3x.opcode_310 as opcode_311
from xdis.opcodes.base import (
    def_op,
    finalize_opcodes,
    init_opdata,
    jrel_op,
    nargs_op,
    rm_op,
    update_pj3,
    varargs_op,
)
from xdis.opcodes.opcode_3x.opcode_311 import opcode_arg_fmt311, opcode_extended_fmt311
from xdis.version_info import PythonImplementation

version_tuple = (3, 11)
python_implementation = PythonImplementation("PyPy")

loc = locals()
init_opdata(loc, opcode_311, version_tuple, is_pypy=True)


# fmt: off
# Changed opcodes
# ----------------------

# Removed from Python 3.11
# ----------------

rm_op(loc,  "GEN_START",              129)
rm_op(loc,  "LIST_EXTEND",            162)
rm_op(loc,  "SET_UPDATE",             163)
rm_op(loc,  "DICT_MERGE",             164)
rm_op(loc,  "DICT_UPDATE",            165)

# Added in PyPy 3.11
# --------------------

def_op(loc, "CHECK_EG_MATCH",                  37,   0, 0)
jrel_op(loc, "PREP_RERAISE_STAR",              88,   1, 0)
def_op(loc, "LIST_EXTEND",                    164,   1, 1)
def_op(loc, "SET_UPDATE",                     165,   2, 1)
def_op(loc, "DICT_MERGE",                     166,   2, 1)
def_op(loc, "DICT_UPDATE",                    167,   2, 1)

loc["hasvargs"].append(202)
nargs_op(loc, "CALL_METHOD_KW",            204, -1, 1)

# Used only in single-mode compilation list-comprehension generators
jrel_op(loc, 'SETUP_EXCEPT',               120,  0,  6, conditional=True)  # ""
varargs_op(loc, "BUILD_LIST_FROM_ARG",     203)
def_op(loc, "LOAD_REVDB_VAR",              205)


# fmt: on

opcode_arg_fmt = opcode_arg_fmt311pypy = opcode_arg_fmt311.copy()
opcode_extended_fmt = opcode_extended_fmt311pypy = opcode_extended_fmt311.copy()

update_pj3(globals(), loc, True)
finalize_opcodes(loc)



# (C) Copyright 2025 by Rocky Bernstein
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
CPython 3.14 bytecode opcodes

This is a like Python 3.14's opcode.py with some classification
of stack usage and information for formatting instructions.
of stack usage.
"""

import xdis.opcodes.opcode_3x.opcode_313 as opcode_313
from xdis.opcodes.base import (
    VARYING_STACK_INT,
    binary_op,
    call_op,
    compare_op,
    const_op,
    cpython_implementation,
    def_op,
    finalize_opcodes,
    free_op,
    init_opdata,
    jrel_op,
    local_op,
    name_op,
    nargs_op,
    store_op,
    unary_op,
    update_pj3,
    varargs_op,
)
from xdis.opcodes.format.extended import extended_format_binary_op

version_tuple = (3, 14)
python_implementation = cpython_implementation

loc = locals()

init_opdata(loc, None, version_tuple)

# extend opcodes to cover pseudo ops
loc["opname"].extend([f"<{i}>" for i in range(256, 267)])
loc["oppop"].extend([0] * 11)
loc["oppush"].extend([0] * 11)

# fmt: off
#            OP NAME                                            OPCODE  POP  PUSH
# ---------------------------------------------------------------------------
def_op(loc,     "CACHE",                                        0,      0,  0)
def_op(loc,     "BINARY_SLICE",                                 1,      3,  1)
def_op(loc,     "BUILD_TEMPLATE",                               2,      2,  1)
local_op(loc,   "BINARY_OP_INPLACE_ADD_UNICODE",                3,      2,  0)
def_op(loc,     "CALL_FUNCTION_EX",                             4,      4,  1)
def_op(loc,     "CHECK_EG_MATCH",                               5,      2,  2)
def_op(loc,     "CHECK_EXC_MATCH",                              6,      2,  2)
def_op(loc,     "CLEANUP_THROW",                                7,      3,  2)
def_op(loc,     "DELETE_SUBSCR",                                8,      2,  0)
def_op(loc,     "END_FOR",                                      9,      1,  0)
def_op(loc,     "END_SEND",                                     10,     2,  1)
def_op(loc,     "EXIT_INIT_CHECK",                              11,     1,  0)
def_op(loc,     "FORMAT_SIMPLE",                                12,     1,  1)
def_op(loc,     "FORMAT_WITH_SPEC",                             13,     2,  1)
def_op(loc,     "GET_AITER",                                    14,     1,  1)
def_op(loc,     "GET_ANEXT",                                    15,     1,  2)
def_op(loc,     "GET_ITER",                                     16,     0,  0)
def_op(loc,     "RESERVED",                                     17,     0,  0)
def_op(loc,     "GET_LEN",                                      18,     1,  2)
def_op(loc,     "GET_YIELD_FROM_ITER",                          19,     1,  1)
def_op(loc,     "INTERPRETER_EXIT",                             20,     1,  0)
def_op(loc,     "LOAD_BUILD_CLASS",                             21,     0,  1)
def_op(loc,     "LOAD_LOCALS",                                  22,     0,  1) # Pushes a reference to the locals of the current scope.
                                                                               # This is not a name op.
nargs_op(loc,   "MAKE_FUNCTION",                                23,     VARYING_STACK_INT,  1)
def_op(loc,     "MATCH_KEYS",                                   24,     2,  3)
def_op(loc,     "MATCH_MAPPING",                                25,     1,  2)
def_op(loc,     "MATCH_SEQUENCE",                               26,     1,  2)
def_op(loc,     "NOP",                                          27,     0,  0)
def_op(loc,     "NOT_TAKEN",                                    28,     0,  0)
def_op(loc,     "POP_EXCEPT",                                   29,     1,  0)
def_op(loc,     "POP_ITER",                                     30,     1,  0)
def_op(loc,     "POP_TOP",                                      31,     1,  0)
def_op(loc,     "PUSH_EXC_INFO",                                32,     1,  2)
def_op(loc,     "PUSH_NULL",                                    33,     0,  1)
def_op(loc,     "RETURN_GENERATOR",                             34,     0,  1)
def_op(loc,     "RETURN_VALUE",                                 35,     1,  1)
def_op(loc,     "SETUP_ANNOTATIONS",                            36,     0,  0)
store_op(loc,   "STORE_SLICE",                                  37,     4,  0)
store_op(loc,   "STORE_SUBSCR",                                 38,     3,  0)
unary_op(loc,   "TO_BOOL",                                      39,     1,  1)
unary_op(loc,   "UNARY_INVERT",                                 40,     1,  1)
unary_op(loc,   "UNARY_NEGATIVE",                               41,     1,  1)
unary_op(loc,   "UNARY_NOT",                                    42,     1,  1)
def_op(loc,     "WITH_EXCEPT_START",                            43,     5,  6)
binary_op(loc,  "BINARY_OP",                                    44,     2,  1)
def_op(loc,     "BUILD_INTERPOLATION",                          45,  VARYING_STACK_INT,  1)  # Either -1 or -2:  pops 2 + (oparg & 1) and pushes result
varargs_op(loc, "BUILD_LIST",                                   46,     -1, 1)  # TOS is count of list items
varargs_op(loc, "BUILD_MAP",                                    47,     0,  1)  # argument is dictionary count to be popped
varargs_op(loc, "BUILD_SET",                                    48,     -1, 1)  # TOS is count of set items
varargs_op(loc, "BUILD_SLICE",                                  49,     -1, 1)  # TOS is slice
varargs_op(loc, "BUILD_STRING",                                 50,     -1, 1)  # TOS is concatenated strings
varargs_op(loc, "BUILD_TUPLE",                                  51,     -1, 1)  # TOS is count of tuple items
call_op(loc,    "CALL",                                         52,     VARYING_STACK_INT, 2)  # pops 2 + oparg; TOS is return value
def_op(loc,     "CALL_INTRINSIC_1",                             53,     1,  1)
def_op(loc,     "CALL_INTRINSIC_2",                             54,     2,  1)
call_op(loc,    "CALL_KW",                                      55,     -3, 1)  # pops 3 + oparg; TOS is return value
compare_op(loc, "COMPARE_OP",                                   56,     2,  1)
binary_op(loc,  "CONTAINS_OP",                                  57,     2,  1)
def_op(loc,     "CONVERT_VALUE",                                58,     1,  1)
def_op(loc,     "COPY",                                         59,     0,  1)
def_op(loc,     "COPY_FREE_VARS",                               60,     0,  0)
name_op(loc,    "DELETE_ATTR",                                  61,     1,  0)
free_op(loc,    "DELETE_DEREF",                                 62,     0,  0)
local_op(loc,   "DELETE_FAST",                                  63,     0,  0)
name_op(loc,    "DELETE_GLOBAL",                                64,     0,  0)
name_op(loc,    "DELETE_NAME",                                  65,     0,  0)
def_op(loc,     "DICT_MERGE",                                   66,     1,  0)
def_op(loc,     "DICT_UPDATE",                                  67,     1,  0)
jrel_op(loc,    "END_ASYNC_FOR",                                68,     2,  0, conditional=True)
def_op(loc,     "EXTENDED_ARG",                                 69,     0,  0)
jrel_op(loc,    "FOR_ITER",                                     70,     0,  1, conditional=True)
def_op(loc,     "GET_AWAITABLE",                                71,     1,  1)
name_op(loc,    "IMPORT_FROM",                                  72,     1,  2)
name_op(loc,    "IMPORT_NAME",                                  73,     2,  1)
compare_op(loc, "IS_OP",                                        74,     2,  1)
jrel_op(loc,    "JUMP_BACKWARD",                                75,     0,  0, conditional=False)
jrel_op(loc,    "JUMP_BACKWARD_NO_INTERRUPT",                   76,     0,  0, conditional=False)
jrel_op(loc,    "JUMP_FORWARD",                                 77,     0,  0, conditional=False)
def_op(loc,     "LIST_APPEND",                                  78,     1,  0)
def_op(loc,     "LIST_EXTEND",                                  79,     1,  0)
name_op(loc,    "LOAD_ATTR",                                    80,     1,  1)  # pops 1 + (oparg & 1)
def_op(loc,     "LOAD_COMMON_CONSTANT",                         81,     0,  1)
const_op(loc,   "LOAD_CONST",                                   82,     0,  1)
local_op(loc,   "LOAD_DEREF",                                   83,     0,  1)
local_op(loc,   "LOAD_FAST",                                    84,     0,  1)
local_op(loc,   "LOAD_FAST_AND_CLEAR",                          85,     0,  1)
local_op(loc,   "LOAD_FAST_BORROW",                             86,     0,  1)
local_op(loc,   "LOAD_FAST_BORROW_LOAD_FAST_BORROW",            87,     0,  2)
local_op(loc,   "LOAD_FAST_CHECK",                              88,     0,  1)
local_op(loc,   "LOAD_FAST_LOAD_FAST",                          89,     0,  2)
free_op(loc,    "LOAD_FROM_DICT_OR_DEREF",                      90,     1,  1)
name_op(loc,    "LOAD_FROM_DICT_OR_GLOBALS",                    91,     1,  1)
name_op(loc,    "LOAD_GLOBAL",                                  92,     0,  1)  # pops 1 + (oparg & 1)
name_op(loc,    "LOAD_NAME",                                    93,     0,  1)
def_op(loc,     "LOAD_SMALL_INT",                               94,     0,  1)
def_op(loc,     "LOAD_SPECIAL",                                 95,     1,  2)
name_op(loc,    "LOAD_SUPER_ATTR",                              96,     3,  1)  # pops 1 + (oparg & 1)
free_op(loc,    "MAKE_CELL",                                    97,     0,  0)
def_op(loc,     "MAP_ADD",                                      98,     2,  0)
def_op(loc,     "MATCH_CLASS",                                  99,     3,  1)
jrel_op(loc,    "POP_JUMP_IF_FALSE",                            100,    1,  0, conditional=True)
jrel_op(loc,    "POP_JUMP_IF_NONE",                             101,    1,  0, conditional=True)
jrel_op(loc,    "POP_JUMP_IF_NOT_NONE",                         102,    1,  0, conditional=True)
jrel_op(loc,    "POP_JUMP_IF_TRUE",                             103,    1,  0, conditional=True)
varargs_op(loc, "RAISE_VARARGS",                                104,    -1, 0)
def_op(loc,     "RERAISE",                                      105,    1,  0)
jrel_op(loc,    "SEND",                                         106,    2,  2, conditional=True)
def_op(loc,     "SET_ADD",                                      107,    1,  0)
def_op(loc,     "SET_FUNCTION_ATTRIBUTE",                       108,    2,  1)
def_op(loc,     "SET_UPDATE",                                   109,    1,  0)
store_op(loc,   "STORE_ATTR",                                   110,    2,  0, is_type="name")
store_op(loc,   "STORE_DEREF",                                  111,    1,  0, is_type="free")
store_op(loc,   "STORE_FAST",                                   112,    1,  0, is_type="local")
local_op(loc,   "STORE_FAST_LOAD_FAST",                         113,    1,  1)
store_op(loc,   "STORE_FAST_STORE_FAST",                        114,    2,  0, is_type="local")
store_op(loc,   "STORE_GLOBAL",                                 115,    1,  0, is_type="name")
store_op(loc,   "STORE_NAME",                                   116,    1,  0, is_type="name")
def_op(loc,     "SWAP",                                         117,    0,  0)
varargs_op(loc, "UNPACK_EX",                                    118,    VARYING_STACK_INT, VARYING_STACK_INT)  # pushes 1 + (oparg & 0xFF) + (oparg >> 8)
varargs_op(loc, "UNPACK_SEQUENCE",                              119,    1, VARYING_STACK_INT)  # unpacks TOS, arg is the count
def_op(loc,     "YIELD_VALUE",                                  120,    1,  1)
def_op(loc,     "RESUME",                                       128,    0,  0)

# Specialized opcodes (>128)
binary_op(loc,  "BINARY_OP_ADD_FLOAT",                          129,    2,  1)
binary_op(loc,  "BINARY_OP_ADD_INT",                            130,    2,  1)
binary_op(loc,  "BINARY_OP_ADD_UNICODE",                        131,    2,  1)
binary_op(loc,  "BINARY_OP_EXTEND",                             132,    2,  1)
binary_op(loc,  "BINARY_OP_MULTIPLY_FLOAT",                     133,    2,  1)
binary_op(loc,  "BINARY_OP_MULTIPLY_INT",                       134,    2,  1)
binary_op(loc,  "BINARY_OP_SUBSCR_DICT",                        135,    2,  1)
binary_op(loc,  "BINARY_OP_SUBSCR_GETITEM",                     136,    2,  0)
binary_op(loc,  "BINARY_OP_SUBSCR_LIST_INT",                    137,    2,  1)
binary_op(loc,  "BINARY_OP_SUBSCR_LIST_SLICE",                  138,    2,  1)
binary_op(loc,  "BINARY_OP_SUBSCR_STR_INT",                     139,    2,  1)
binary_op(loc,  "BINARY_OP_SUBSCR_TUPLE_INT",                   140,    2,  1)
binary_op(loc,  "BINARY_OP_SUBTRACT_FLOAT",                     141,    2,  1)
binary_op(loc,  "BINARY_OP_SUBTRACT_INT",                       142,    2,  1)
call_op(loc,    "CALL_ALLOC_AND_ENTER_INIT",                    143,    -2, 0)  # pops 2 + oparg
call_op(loc,    "CALL_BOUND_METHOD_EXACT_ARGS",                 144,    -2, 0)  # pops 2 + oparg
call_op(loc,    "CALL_BOUND_METHOD_GENERAL",                    145,    -2, 0)  # pops 2 + oparg
call_op(loc,    "CALL_BUILTIN_CLASS",                           146,    -2, 1)  # pops 2 + oparg
call_op(loc,    "CALL_BUILTIN_FAST",                            147,    -2, 1)  # pops 2 + oparg
call_op(loc,    "CALL_BUILTIN_FAST_WITH_KEYWORDS",              148,    -2, 1)  # pops 2 + oparg
call_op(loc,    "CALL_BUILTIN_O",                               149,    -2, 1)  # pops 2 + oparg
def_op(loc,     "CALL_ISINSTANCE",                              150,    4,  1)
call_op(loc,    "CALL_KW_BOUND_METHOD",                         151,    -3, 0)  # pops 3 + oparg
call_op(loc,    "CALL_KW_NON_PY",                               152,    -3, 1)  # pops 3 + oparg
call_op(loc,    "CALL_KW_PY",                                   153,    -3, 0)  # pops 3 + oparg
def_op(loc,     "CALL_LEN",                                     154,    3,  1)
def_op(loc,     "CALL_LIST_APPEND",                             155,    3,  0)
call_op(loc,    "CALL_METHOD_DESCRIPTOR_FAST",                  156,    -2, 1)  # pops 2 + oparg
call_op(loc,    "CALL_METHOD_DESCRIPTOR_FAST_WITH_KEYWORDS",    157,    -2, 1)  # pops 2 + oparg
call_op(loc,    "CALL_METHOD_DESCRIPTOR_NOARGS",                158,    -2, 1)  # pops 2 + oparg
call_op(loc,    "CALL_METHOD_DESCRIPTOR_O",                     159,    -2, 1)  # pops 2 + oparg
call_op(loc,    "CALL_NON_PY_GENERAL",                          160,    -2, 1)  # pops 2 + oparg
call_op(loc,    "CALL_PY_EXACT_ARGS",                           161,    -2, 0)  # pops 2 + oparg
call_op(loc,    "CALL_PY_GENERAL",                              162,    -2, 0)  # pops 2 + oparg
def_op(loc,     "CALL_STR_1",                                   163,    3,  1)
def_op(loc,     "CALL_TUPLE_1",                                 164,    3,  1)
def_op(loc,     "CALL_TYPE_1",                                  165,    3,  1)
compare_op(loc, "COMPARE_OP_FLOAT",                             166,    2,  1)
compare_op(loc, "COMPARE_OP_INT",                               167,    2,  1)
compare_op(loc, "COMPARE_OP_STR",                               168,    2,  1)
def_op(loc,     "CONTAINS_OP_DICT",                             169,    2,  1)
def_op(loc,     "CONTAINS_OP_SET",                              170,    2,  1)
jrel_op(loc,    "FOR_ITER_GEN",                                 171,    2,  2, conditional=True)
jrel_op(loc,    "FOR_ITER_LIST",                                172,    2,  3, conditional=True)
jrel_op(loc,    "FOR_ITER_RANGE",                               173,    2,  3, conditional=True)
jrel_op(loc,    "FOR_ITER_TUPLE",                               174,    2,  3, conditional=True)
jrel_op(loc,    "JUMP_BACKWARD_JIT",                            175,    0,  0, conditional=False)
jrel_op(loc,    "JUMP_BACKWARD_NO_JIT",                         176,    0,  0, conditional=False)
def_op(loc,     "LOAD_ATTR_CLASS",                              177,    1,  1)  # pushes 1 + (oparg & 1)
def_op(loc,     "LOAD_ATTR_CLASS_WITH_METACLASS_CHECK",         178,    1,  1)  # pushes 1 + (oparg & 1)
name_op(loc,    "LOAD_ATTR_GETATTRIBUTE_OVERRIDDEN",            179,    1,  1)
def_op(loc,     "LOAD_ATTR_INSTANCE_VALUE",                     180,    1,  1)  # pushes 1 + (oparg & 1)
def_op(loc,     "LOAD_ATTR_METHOD_LAZY_DICT",                   181,    1,  2)
def_op(loc,     "LOAD_ATTR_METHOD_NO_DICT",                     182,    1,  2)
def_op(loc,     "LOAD_ATTR_METHOD_WITH_VALUES",                 183,    1,  2)
def_op(loc,     "LOAD_ATTR_MODULE",                             184,    1,  1)  # pushes 1 + (oparg & 1)
def_op(loc,     "LOAD_ATTR_NONDESCRIPTOR_NO_DICT",              185,    1,  1)
def_op(loc,     "LOAD_ATTR_NONDESCRIPTOR_WITH_VALUES",          186,    1,  1)
def_op(loc,     "LOAD_ATTR_PROPERTY",                           187,    1,  0)
def_op(loc,     "LOAD_ATTR_SLOT",                               188,    1,  1)  # pushes 1 + (oparg & 1)
name_op(loc,    "LOAD_ATTR_WITH_HINT",                          189,    1,  1)  # pushes 1 + (oparg & 1)
def_op(loc,     "LOAD_GLOBAL_BUILTIN",                          190,    0,  1)  # pushes 1 + (oparg & 1)
def_op(loc,     "LOAD_GLOBAL_MODULE",                           191,    0,  1)  # pushes 1 + (oparg & 1)
name_op(loc,    "LOAD_SUPER_ATTR_ATTR",                         192,    3,  1)
name_op(loc,    "LOAD_SUPER_ATTR_METHOD",                       193,    3,  2)
def_op(loc,     "RESUME_CHECK",                                 194,    0,  0)
def_op(loc,     "SEND_GEN",                                     195,    2,  1)
def_op(loc,     "STORE_ATTR_INSTANCE_VALUE",                    196,    2,  0)
def_op(loc,     "STORE_ATTR_SLOT",                              197,    2,  0)
store_op(loc,   "STORE_ATTR_WITH_HINT",                         198,    2,  0, is_type="name")
def_op(loc,     "STORE_SUBSCR_DICT",                            199,    3,  0)
def_op(loc,     "STORE_SUBSCR_LIST_INT",                        200,    3,  0)
def_op(loc,     "TO_BOOL_ALWAYS_TRUE",                          201,    1,  1)
def_op(loc,     "TO_BOOL_BOOL",                                 202,    1,  1)
def_op(loc,     "TO_BOOL_INT",                                  203,    1,  1)
def_op(loc,     "TO_BOOL_LIST",                                 204,    1,  1)
def_op(loc,     "TO_BOOL_NONE",                                 205,    1,  1)
def_op(loc,     "TO_BOOL_STR",                                  206,    1,  1)
varargs_op(loc, "UNPACK_SEQUENCE_LIST",                         207,    1,  -1)
varargs_op(loc, "UNPACK_SEQUENCE_TUPLE",                        208,    1,  -1)
def_op(loc,     "UNPACK_SEQUENCE_TWO_TUPLE",                    209,    1,  2)
def_op(loc,     "INSTRUMENTED_END_FOR",                         233,    3,  2)
def_op(loc,     "INSTRUMENTED_POP_ITER",                        234,    1,  0)
def_op(loc,     "INSTRUMENTED_END_SEND",                        235,    2,  1)
jrel_op(loc,    "INSTRUMENTED_FOR_ITER",                        236,    2,  1, conditional=True)
def_op(loc,     "INSTRUMENTED_INSTRUCTION",                     237,    0,  1)
jrel_op(loc,    "INSTRUMENTED_JUMP_FORWARD",                    238,    0,  0, conditional=False)
def_op(loc,     "INSTRUMENTED_NOT_TAKEN",                       239,    0,  0)
jrel_op(loc,    "INSTRUMENTED_POP_JUMP_IF_TRUE",                240,    0,  0, conditional=True)  # dunno why it's not 1, 0.
jrel_op(loc,    "INSTRUMENTED_POP_JUMP_IF_FALSE",               241,    1,  0, conditional=True)
jrel_op(loc,    "INSTRUMENTED_POP_JUMP_IF_NONE",                242,    1,  0, conditional=True)
jrel_op(loc,    "INSTRUMENTED_POP_JUMP_IF_NOT_NONE",            243,    1,  0, conditional=True)
def_op(loc,     "INSTRUMENTED_RESUME",                          244,    1,  0)  # dunno why it's not 0, 0.
def_op(loc,     "INSTRUMENTED_RETURN_VALUE",                    245,    1,  1)
def_op(loc,     "INSTRUMENTED_YIELD_VALUE",                     246,    1,  1)
jrel_op(loc,    "INSTRUMENTED_END_ASYNC_FOR",                   247,    0,  0, conditional=True) # dunnow why it's not 2, 0
name_op(loc,    "INSTRUMENTED_LOAD_SUPER_ATTR",                 248,    VARYING_STACK_INT,  1)  # pushes 1 + (oparg & 1)
call_op(loc,    "INSTRUMENTED_CALL",                            249,    -2, 1)
call_op(loc,    "INSTRUMENTED_CALL_KW",                         250,    -3, 1)
call_op(loc,    "INSTRUMENTED_CALL_FUNCTION_EX",                251,    4,  1)
jrel_op(loc,    "INSTRUMENTED_JUMP_BACKWARD",                   252,    0,  0, conditional=False)
def_op(loc,     "INSTRUMENTED_LINE",                            253,    0,  0)
def_op(loc,     "ENTER_EXECUTOR",                               254,    0,  0)
def_op(loc,     "TRACE_RECORD",                                 255,    0,  0)
def_op(loc,     "ANNOTATIONS_PLACEHOLDER",                      256,    0,  0)
jrel_op(loc,    "JUMP",                                         257,    0,  0, conditional=False)
jrel_op(loc,    "JUMP_IF_FALSE",                                258,    1,  1, conditional=True)
jrel_op(loc,    "JUMP_IF_TRUE",                                 259,    1,  1, conditional=True)
jrel_op(loc,    "JUMP_NO_INTERRUPT",                            260,    0,  0, conditional=False)
local_op(loc,   "LOAD_CLOSURE",                                 261,    0,  1)
def_op(loc,     "POP_BLOCK",                                    262,    0,  0)
def_op(loc,     "SETUP_CLEANUP",                                263,    0,  2)
def_op(loc,     "SETUP_FINALLY",                                264,    0,  1)
def_op(loc,     "SETUP_WITH",                                   265,    0,  1)
def_op(loc,     "STORE_FAST_MAYBE_NULL",                        266,    1,  0)

# ops >= 44 have args
HAVE_ARGUMENT = 44

loc["hasarg"] = [44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 128, 143, 144, 145, 146, 147, 148, 149, 151, 152, 153, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 195, 198, 207, 208, 209, 236, 238, 240, 241, 242, 243, 244, 246, 247, 248, 249, 250, 252, 254, 255, 257, 258, 259, 260, 261, 263, 264, 265, 266]
# hasconst table populated by const_op definitions
# hasname table populated by name_op and store_op definitions
# jrel table populated by jrel_op definitions
loc["hasjabs"] = []
loc["hasjump"] = loc["hasjrel"]
# hasfree table populated by free_op and store_op definitions
# haslocal table populated by local_op and store_op definitions
loc["hasexc"] = [263, 264, 265]

# fmt: on

### update formatting

_nb_ops = [
    ("NB_ADD", "+"),
    ("NB_AND", "&"),
    ("NB_FLOOR_DIVIDE", "//"),
    ("NB_LSHIFT", "<<"),
    ("NB_MATRIX_MULTIPLY", "@"),
    ("NB_MULTIPLY", "*"),
    ("NB_REMAINDER", "%"),
    ("NB_OR", "|"),
    ("NB_POWER", "**"),
    ("NB_RSHIFT", ">>"),
    ("NB_SUBTRACT", "-"),
    ("NB_TRUE_DIVIDE", "/"),
    ("NB_XOR", "^"),
    ("NB_INPLACE_ADD", "+="),
    ("NB_INPLACE_AND", "&="),
    ("NB_INPLACE_FLOOR_DIVIDE", "//="),
    ("NB_INPLACE_LSHIFT", "<<="),
    ("NB_INPLACE_MATRIX_MULTIPLY", "@="),
    ("NB_INPLACE_MULTIPLY", "*="),
    ("NB_INPLACE_REMAINDER", "%="),
    ("NB_INPLACE_OR", "|="),
    ("NB_INPLACE_POWER", "**="),
    ("NB_INPLACE_RSHIFT", ">>="),
    ("NB_INPLACE_SUBTRACT", "-="),
    ("NB_INPLACE_TRUE_DIVIDE", "/="),
    ("NB_INPLACE_XOR", "^="),
    ("NB_SUBSCR", "[]"),
]


def format_BINARY_OP_314(arg: int):
    return _nb_ops[arg][1]


def extended_BINARY_OP_314(opc, instructions):
    opname = _nb_ops[instructions[0].argval][1]

    if opname == "%":
        # Make sure to escape % below.
        opname = "%%"

    fmt_str = "%s[%s]" if opname == "[]" else f"%s {opname} %s"
    return extended_format_binary_op(opc, instructions, fmt_str)


_common_constants = ("AssertionError", "NotImplementedError", "tuple", "all", "any")


def format_LOAD_COMMON_CONSTANT_314(arg: int):
    return _common_constants[arg]
    return _common_constants[arg]


opcode_arg_fmt = opcode_arg_fmt314 = {
    **opcode_313.opcode_arg_fmt313,
    **{"BINARY_OP": format_BINARY_OP_314},
    **{"LOAD_COMMON_CONSTANT": format_LOAD_COMMON_CONSTANT_314},
}

opcode_extended_fmt = opcode_extended_fmt314 = {
    **opcode_313.opcode_extended_fmt313,
    **{"BINARY_OP": extended_BINARY_OP_314},
}

# CALL_FUNCTION_EX no longer takes an argument in 3.14, so it no longer needs to be formatted
del opcode_arg_fmt["CALL_FUNCTION_EX"]

findlinestarts = opcode_313.findlinestarts_313

update_pj3(globals(), loc)
finalize_opcodes(loc)



# (C) Copyright 2018-2025 by Rocky Bernstein
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Internal Python object serialization

This module contains functions that can read and write Python values
in a binary format. The format is specific to Python, but independent
of machine architecture issues (e.g., you can write a Python value to
a file on a PC, transport the file to a Sun, and read it back
there). Details of the format may change between Python versions.

"""

import struct
import types
from sys import intern
from types import CodeType, EllipsisType
from typing import Any, Dict, Optional, Set, Union

from xdis.codetype import Code2, Code3, Code15
from xdis.unmarshal import (
    FLAG_REF,
    TYPE_ASCII,
    TYPE_BINARY_COMPLEX,
    TYPE_BINARY_FLOAT,
    TYPE_CODE,
    TYPE_COMPLEX,
    TYPE_DICT,
    TYPE_ELLIPSIS,
    TYPE_FALSE,
    TYPE_FLOAT,
    TYPE_FROZENSET,
    TYPE_INT,
    TYPE_INT64,
    TYPE_INTERNED,
    TYPE_LIST,
    TYPE_LONG,
    TYPE_NONE,
    TYPE_NULL,
    TYPE_REF,
    TYPE_SET,
    TYPE_SHORT_ASCII,
    TYPE_SHORT_ASCII_INTERNED,
    TYPE_SMALL_TUPLE,
    TYPE_STOPITER,
    TYPE_STRING,
    TYPE_STRINGREF,
    TYPE_TRUE,
    TYPE_TUPLE,
    TYPE_UNICODE,
    long,
)
from xdis.version_info import PYTHON_VERSION_TRIPLE, version_tuple_to_str

# NOTE: This module is used in the Python3 interpreter, but also by
# the "sandboxed" process.  It must work for Python2 as well.

try:
    from __pypy__ import builtinify

except ImportError:

    def builtinify(f):
        return f


class _Marshaller:
    """Python marshalling routine that runs in Python 2 and Python 3.
    We also extend to allow for xdis Code15, Code2, and Code3 types and instances.
    """

    dispatch = {}

    def __init__(
        self,
        writefunc,
        python_version: tuple,
        is_pypy: Optional[bool] = None,
        collection_order={},
        reference_objects=set(),
    ) -> None:
        self._write = writefunc
        self.collection_order = collection_order
        self.intern_objects: Dict[Any, int] = {}
        self.intern_consts: Dict[Any, int] = {}
        self.is_pypy = is_pypy
        self.python_version = python_version
        self.reference_objects = reference_objects

    def dump(self, x, flag_ref: int = 0) -> None:
        if (
            isinstance(x, types.CodeType)
            and PYTHON_VERSION_TRIPLE[:2] != self.python_version[:2]
        ):
            raise RuntimeError(
                "code type passed for version %s but we are running version %s"
                % (version_tuple_to_str(), self.python_version)
            )
        try:
            self.dispatch[type(x)](self, x, flag_ref)
        except KeyError:
            if isinstance(x, Code3):
                self.dispatch[Code3](self, x)
                return
            elif isinstance(x, Code2):
                self.dispatch[Code2](self, x)
                return
            elif isinstance(x, Code15):
                self.dispatch[Code15](self, x)
                return
            else:
                for tp in type(x).mro():
                    func = self.dispatch.get(tp)
                    if func:
                        break
                else:
                    raise ValueError("unmarshallable object")
            func(self, x)

    # FIXME: Handle interned versions of dump_ascii, dump_short_ascii
    def dump_ascii(self, s: str) -> None:
        self._write(TYPE_ASCII)
        self.w_long(len(s))
        self._write(s)

    dispatch[TYPE_ASCII] = dump_ascii

    def dump_binary_complex(self, x) -> None:
        write = self._write
        write(TYPE_BINARY_COMPLEX)
        write(struct.pack("<d", x.real))
        write(struct.pack("<d", x.imag))

    dispatch[TYPE_BINARY_COMPLEX] = dump_binary_complex

    def dump_binary_float(self, x) -> None:
        write = self._write
        write(TYPE_BINARY_FLOAT)
        write(struct.pack("<d", x))

    def dump_bool(self, x) -> None:
        if x:
            self._write(TYPE_TRUE)
        else:
            self._write(TYPE_FALSE)

    dispatch[bool] = dump_bool

    def dump_code15(self, x) -> None:
        # Careful here: many Python 2 code objects are strings,
        # but Python 3 marshaling, by default, will dump strings as
        # unicode. Force marsaling this type as string.

        self._write(TYPE_CODE)
        self.w_short(x.co_argcount)
        self.w_short(x.co_nlocals)
        self.w_short(x.co_stacksize)
        self.w_short(x.co_flags)
        self.dump_string(x.co_code)

        # If running in a Python3 interpreter, some constants will get
        # converted from string to unicode. For now, let's see if
        # that's okay.
        self.dump(x.co_consts)

        # The tuple "names" in Python 1.x must have string entries
        self._write(TYPE_TUPLE)
        self.w_long(len(x.co_names))
        for name in x.co_names:
            self.dump_string(name)

        # The tuple "varnames" in Python 1.x also must have string entries
        self._write(TYPE_TUPLE)
        self.w_long(len(x.co_varnames))
        for name in x.co_varnames:
            self.dump_string(name)

        self.dump_string(x.co_filename)
        self.dump_string(x.co_name)
        self.w_long(x.co_firstlineno)
        self.dump_string(x.co_lnotab)
        return

    dispatch[Code15] = dump_code15

    def dump_code2(self, x, _) -> None:
        # Careful here: many Python 2 code objects are strings,
        # but Python 3 marshaling, by default, will dump strings as
        # unicode. Force marsaling this type as string.

        self._write(TYPE_CODE)
        self.w_long(x.co_argcount)
        self.w_long(x.co_nlocals)
        self.w_long(x.co_stacksize)
        self.w_long(x.co_flags)
        self.dump_string(x.co_code)

        # If running in a Python3 interpreter, some constants will get
        # converted from string to unicode. For now, let's see if
        # that's okay.
        self.dump(x.co_consts)

        # The tuple "names" in Python2 must have string entries
        self._write(TYPE_TUPLE)
        self.w_long(len(x.co_names))
        for name in x.co_names:
            self.dump_string(name)

        # The tuple "varnames" in Python2 also must have string entries
        self._write(TYPE_TUPLE)
        self.w_long(len(x.co_varnames))
        for name in x.co_varnames:
            self.dump_string(name)

        self.dump(x.co_freevars)
        self.dump(x.co_cellvars)
        self.dump_string(x.co_filename)
        self.dump_string(x.co_name)
        self.w_long(x.co_firstlineno)
        self.dump_string(x.co_lnotab)
        return

    dispatch[Code2] = dump_code2

    # FIXME: will probably have to adjust similar to how we
    # adjusted dump_code2
    def dump_code3(self, code, flag_ref: int = 0) -> None:
        if flag_ref:
            self._write(chr(ord(TYPE_CODE) | flag_ref))

            # The way marshal works for 3.4 (up to ....?)
            # The first object is always None. Supposedly that
            # object is to be filled in by the code at the end,
            # but by filling in "None" first, that gets used should
            # there be other instances of None. Seems like a bug,
            # but this is Python after all
            self.intern_consts[None] = 0

            # This is probably wrong. We are off by one in
            # intern_objects.
            self.intern_objects[code] = len(self.intern_objects)

        else:
            self._write(TYPE_CODE)

        self.w_long(code.co_argcount)
        if hasattr(code, "co_posonlyargcount"):
            self.w_long(code.co_posonlyargcount)
        self.w_long(code.co_kwonlyargcount)
        if self.python_version < (3, 11):
            self.w_long(code.co_nlocals)
        self.w_long(code.co_stacksize)
        self.w_long(code.co_flags)
        self.dump(code.co_code)
        self.dump(code.co_consts, flag_ref)
        self.dump_names(code.co_names, flag_ref)
        self.dump_names(code.co_varnames, flag_ref)
        self.dump_names(code.co_freevars, flag_ref)
        self.dump_names(code.co_cellvars, flag_ref)
        self.dump_filename(code.co_filename, flag_ref)
        self.dump_name(code.co_name, flag_ref)
        self.w_long(code.co_firstlineno)

        # 3.10 and greater uses co_linetable.
        linetable = code.co_linetable if hasattr(code, "co_linetable") else code.co_lnotab
        self.dump_linetable(linetable)

    dispatch[Code3] = dump_code3

    # FIXME: this is wrong.
    try:
        dispatch[types.CodeType] = dump_code3
    except NameError:
        pass

    def dump_collection(self, type_code: str, bag: Union[frozenset, set, dict]) -> None:
        """
        Save marshalled version of frozenset fs.
        Use self.collection_order, to ensure that the order
        or set elements that may have appeared from unmarshalling the appears
        the same way. This helps roundtrip checking, among possibly other things.
        """
        self._write(type_code)
        self.w_long(len(bag))
        collection = self.collection_order.get(bag, bag)
        for each in collection:
            self.dump(each)

    def dump_complex(self, x, _) -> None:
        write = self._write
        write(TYPE_COMPLEX)
        s = repr(x.real)
        write(chr(len(s)))
        write(s)
        s = repr(x.imag)
        write(chr(len(s)))
        write(s)

    try:
        dispatch[complex] = dump_complex
    except NameError:
        pass

    def dump_dict(self, x) -> None:
        self._write(TYPE_DICT)
        for key, value in x.items():
            self.dump(key)
            self.dump(value)
        self._write(TYPE_NULL)

    dispatch[dict] = dump_dict

    def dump_ellipsis(self, _) -> None:
        self._write(TYPE_ELLIPSIS)

    try:
        dispatch[type(Ellipsis)] = dump_ellipsis
    except NameError:
        pass

    def dump_float(self, x) -> None:
        write = self._write
        write(TYPE_FLOAT)
        s = repr(x)
        write(chr(len(s)))
        write(s)

    dispatch[float] = dump_float
    dispatch[TYPE_BINARY_FLOAT] = dump_float

    def dump_filename(self, path, flag_ref: int = 0) -> None:
        if flag_ref:
            # After Python 3.4 which adds the ref flag and ASCII marshal types..
            if len(path) < 256:
                self.dump_short_ascii(path)
            else:
                self.dump_ascii(path)
        else:
            # Python 3.0 .. 3.4...
            self.dump_unicode(path)

    def dump_frozenset(self, fs: frozenset, flag_ref: int = 0) -> None:
        """
        Save marshalled version of frozenset fs.
        """
        self.dump_collection(TYPE_FROZENSET, fs)

    try:
        dispatch[frozenset] = dump_frozenset
    except NameError:
        pass

    def dump_linetable(self, s) -> None:
        type_code = TYPE_STRING if self.python_version < (3, 5) else TYPE_UNICODE
        self._write(type_code)
        self.w_long(len(s))
        self._write(s)

    def w_long(self, x: int) -> None:
        a = chr(x & 0xFF)
        x >>= 8
        b = chr(x & 0xFF)
        x >>= 8
        c = chr(x & 0xFF)
        x >>= 8
        d = chr(x & 0xFF)
        self._write(a + b + c + d)

    def w_long64(self, x) -> None:
        self.w_long(x)
        self.w_long(x >> 32)

    def dump_name(self, name: str, flag_ref: int) -> None:
        if flag_ref:
            if len(name) < 256:
                self.dump_short_ascii_interned(name)
            else:
                self.dump_ascii(name)
        else:
            self.dump_unicode(name)


    def dump_names(self, names, flag_ref: int) -> None:
        n = len(names)
        if flag_ref:

            # We have reference objects. Has "names" already been seen as a reference object?
            if names in self.intern_objects:
                # It has, so just write the reference and return.
                self._write(TYPE_REF)
                self.w_long(self.intern_objects[names])
                return

            if n < 256:
                is_reference = names in self.reference_objects
                type_code = (
                    chr(ord(TYPE_SMALL_TUPLE) | FLAG_REF)
                    if is_reference
                    else TYPE_SMALL_TUPLE
                )
                self._write(type_code)
                self._write(chr(n))
                if is_reference:
                    self.intern_objects[names] = len(self.intern_objects)

            else:
                self._write(chr(ord(TYPE_TUPLE) | FLAG_REF))
                self.w_long(n)

            for name in names:
                self.dump_short_ascii_interned(name)
        else:
            self._write(TYPE_TUPLE)
            self.w_long(n)
            for name in names:
                self.dump(name)

    def dump_none(self, _, flag_ref: int) -> None:
        self._write(TYPE_NONE)
        # In Python 3.4 .. ? None appears always as the first
        # constant.
        if flag_ref and self.intern_consts.get(None, -1) == -1:
            self.intern_consts[None] = len(self.intern_consts)

    dispatch[type(None)] = dump_none

    def dump_int(self, value: int, flag_ref: int = 0) -> None:
        if flag_ref:
            if (ref := self.intern_consts.get(value, None)) is not None:
                self.dump_ref(ref)
                return
            n = len(self.intern_consts)
            self.intern_consts[value] = n

        y = value >> 31
        if y and y != -1:
            self._write(chr(ord(TYPE_INT64) | flag_ref))
            self.w_long64(value)
        else:
            self._write(chr(ord(TYPE_INT) | flag_ref))
            self.w_long(value)

    dispatch[int] = dump_int

    def dump_list(self, x) -> None:
        self._write(TYPE_LIST)
        self.w_long(len(x))
        for item in x:
            self.dump(item)

    dispatch[list] = dump_list

    def dump_long(self, x) -> None:
        self._write(TYPE_LONG)
        sign = 1
        if x < 0:
            sign = -1
            x = -x
        digits = []
        while x:
            digits.append(x & 0x7FFF)
            x = x >> 15
        self.w_long(len(digits) * sign)
        for d in digits:
            self.w_short(d)

    try:
        long
    except NameError:
        dispatch[int] = dump_long
    else:
        dispatch[long] = dump_long  # noqa

    def dump_ref(self, ref: int) -> None:
        self._write(TYPE_REF)
        self.w_long(ref)

    def dump_set(self, s: Set[Any]) -> None:
        """
        Save marshalled version of set s.
        """
        self.dump_collection(TYPE_SET, s)

    dispatch[set] = dump_set

    def w_short(self, x: int) -> None:
        self._write(chr(x & 0xFF))
        self._write(chr((x >> 8) & 0xFF))

    def dump_short_ascii(self, short_ascii: str) -> None:
        self._write(chr(ord(TYPE_SHORT_ASCII) | FLAG_REF))
        # FIXME: check len(x)?
        self._write(chr(len(short_ascii)))
        self._write(short_ascii)

    dispatch[TYPE_SHORT_ASCII] = dump_short_ascii

    def dump_short_ascii_interned(self, short_ascii: str) -> None:
        """
        Used when the length of an ASCII string is less than 255
        characters. This is used in Python 3.4 and later.
        """
        if (ref := self.intern_consts.get(short_ascii, None)) is not None:
            self.dump_ref(ref)
            return

        self._write(chr(ord(TYPE_SHORT_ASCII_INTERNED) | FLAG_REF))
        self._write(chr(len(short_ascii)))
        self._write(short_ascii)
        n = len(self.intern_objects)
        self.intern_objects[short_ascii] = n

    dispatch[TYPE_ASCII] = dump_ascii

    def dump_small_tuple(self, tuple_value, flag_ref: int) -> None:
        """
        Used when the length of a tuple has is less than 255
        items. This is used in Python 3.4 and later.
        """
        if flag_ref:
            if (ref := self.intern_objects.get(tuple_value, None)) is not None:
                self.dump_ref(ref)
                return
            n = len(self.intern_objects)
            self.intern_objects[tuple_value] = n

        self._write(TYPE_SMALL_TUPLE)
        self._write(chr(len(tuple_value)))
        for item in tuple_value:
            self.dump(item, FLAG_REF)

    dispatch[TYPE_SMALL_TUPLE] = dump_small_tuple

    def dump_stopiter(self, x) -> None:
        if x is not StopIteration:
            raise ValueError("unmarshallable object")
        self._write(TYPE_STOPITER)

    dispatch[type(StopIteration)] = dump_stopiter

    def dump_string(self, s, flag_ref: int = 0) -> None:
        # Python 3.11 seems to add the object ref flag bit for strings.
        if self.python_version >= (3, 11):
            type_code = chr(ord(TYPE_STRING) | flag_ref)
        if (3, 0) <= self.python_version < (3, 11):
            type_code = TYPE_STRING
        else:
            # Python 2.x.
            # FIXME: save string somewhere if it isn't in string table.
            type_code = TYPE_INTERNED if s in self.reference_objects else TYPE_STRING

        self._write(type_code)
        self.w_long(len(s))
        self._write(s)

    dispatch[bytes] = dump_string
    dispatch[bytearray] = dump_string

    def dump_tuple(self, tuple_object: tuple, flag_ref: int = 0) -> None:

        n = len(tuple_object)
        if self.python_version >= (3, 4) and n < 256:
            self.dump_small_tuple(tuple_object, flag_ref)
            return

        type_code = TYPE_TUPLE
        self._write(type_code)
        self.w_long(len(tuple_object))
        for item in tuple_object:
            self.dump(item)

    dispatch[tuple] = dump_tuple
    dispatch[TYPE_TUPLE] = dump_tuple
    dispatch[TYPE_LIST] = dump_tuple

    def dump_unicode(self, s, flag_ref: int = 0) -> None:
        if self.python_version < (2, 0):
            type_code = TYPE_STRING
        elif (2, 0) <= self.python_version < (3, 0):
            # FIXME: probably need to save string somewhere
            # if it isn't in string table.
            type_code = TYPE_INTERNED if s in self.reference_objects else TYPE_STRING
        else:
            type_code = TYPE_UNICODE

        if flag_ref:
            if (ref := self.intern_objects.get(s, None)) is not None:
                self.dump_ref(ref)
                return
            n = len(self.intern_objects)
            self.intern_objects[s] = n
            type_code = chr(ord(type_code) | flag_ref)

        self._write(type_code)
        self.w_long(len(s))
        self._write(s)

    try:
        unicode
    except NameError:
        dispatch[str] = dump_unicode
    else:
        dispatch[unicode] = dump_unicode  # noqa



class _NULL:
    pass


class _StringBuffer:
    def __init__(self, value) -> None:
        self.bufstr = value
        self.bufpos = 0

    def read(self, n):
        pos = self.bufpos
        newpos = pos + n
        ret = self.bufstr[pos:newpos]
        self.bufpos = newpos
        return ret


class _Unmarshaller:
    dispatch = {}

    def __init__(self, readfunc, python_version: bool, is_pypy: bool) -> None:
        self._read = readfunc
        self._stringtable = []
        self.is_pypy = is_pypy
        self.python_version = python_version

    def load(self):
        c = self._read(1)
        if not c:
            raise EOFError
        try:
            return self.dispatch[c](self)
        except KeyError:
            raise ValueError("bad marshal code: %s (%s)" % (c, c))

    def r_byte(self):
        return self._read(1)

    def r_long(self):
        s = self._read(4)
        a = s[0]
        b = s[1]
        c = s[2]
        d = s[3]
        x = a | (b << 8) | (c << 16) | (d << 24)
        if d & 0x80 and x > 0:
            x = -((1 << 32) - x)
            return int(x)
        else:
            return x

    dispatch[TYPE_INT] = r_long

    def r_long64(self):
        a = self._read(1)
        b = self._read(1)
        c = self._read(1)
        d = self._read(1)
        e = self._read(1)
        f = self._read(1)
        g = self._read(1)
        h = self._read(1)
        x = a | (b << 8) | (c << 16) | (d << 24)
        x = x | (e << 32) | (f << 40) | (g << 48) | (h << 56)
        if h & 0x80 and x > 0:
            x = -((1 << 64) - x)
        return x

    dispatch[TYPE_INT64] = r_long64

    def r_short(self):
        lo = self._read(1)
        hi = self._read(1)
        x = lo | (hi << 8)
        if x & 0x8000:
            x = x - 0x10000
        return x

    def load_null(self) -> type[_NULL]:
        return _NULL

    dispatch[TYPE_NULL] = load_null

    def load_ascii(self):
        return self.r_byte()

    dispatch[TYPE_ASCII] = load_ascii

    def load_binary_float(self) -> float:
        f = self._read(8)
        return float(struct.unpack("<d", f)[0])

    dispatch[TYPE_BINARY_FLOAT] = load_binary_float

    # FIXME: GO over fo PYPY
    def load_code(self) -> Code2 | Code3 | CodeType:
        argcount = self.r_long()
        if self.python_version and self.python_version >= (3, 0):
            is_python3 = True
            kwonlyargcount = self.r_long()
        else:
            is_python3 = False
        nlocals = self.r_long()
        stacksize = self.r_long()
        flags = self.r_long()
        code = self.load()
        consts = self.load()
        names = self.load()
        varnames = self.load()
        freevars = self.load()
        cellvars = self.load()
        filename = self.load()
        name = self.load()
        firstlineno = self.r_long()
        lnotab = self.load()
        if is_python3:
            return types.CodeType(
                argcount,
                kwonlyargcount,
                nlocals,
                stacksize,
                flags,
                code,
                consts,
                names,
                varnames,
                filename,
                name,
                firstlineno,
                lnotab,
                freevars,
                cellvars,
            )
        else:
            return Code2(
                argcount,
                nlocals,
                stacksize,
                flags,
                code,
                consts,
                names,
                varnames,
                filename,
                name,
                firstlineno,
                lnotab,
                freevars,
                cellvars,
            )

    dispatch[TYPE_CODE] = load_code

    def load_complex(self) -> complex:
        n = self._read(1)
        s = self._read(n)
        real = float(s)
        n = self._read(1)
        s = self._read(n)
        imag = float(s)
        return complex(real, imag)

    dispatch[TYPE_COMPLEX] = load_complex

    def load_dict(self):
        d = {}
        while 1:
            key = self.load()
            if key is _NULL:
                break
            value = self.load()
            d[key] = value
        return d

    dispatch[TYPE_DICT] = load_dict

    def load_ellipsis(self) -> EllipsisType:
        return Ellipsis

    dispatch[TYPE_ELLIPSIS] = load_ellipsis

    def load_false(self) -> bool:
        return False

    dispatch[TYPE_FALSE] = load_false

    def load_float(self) -> float:
        n = self._read(1)
        s = self._read(n)
        return float(s)

    dispatch[TYPE_FLOAT] = load_float

    def load_frozenset(self):
        n = self.r_long()
        args = [self.load() for i in range(n)]
        return frozenset(args)

    dispatch[TYPE_FROZENSET] = load_frozenset

    def load_interned(self) -> str:
        n = self.r_long()
        ret = intern(self._read(n))
        self._stringtable.append(ret)
        return ret

    dispatch[TYPE_INTERNED] = load_interned

    def load_list(self):
        n = self.r_long()
        list = [self.load() for i in range(n)]
        return list

    dispatch[TYPE_LIST] = load_list

    def load_long(self):
        size = self.r_long()
        sign = 1
        if size < 0:
            sign = -1
            size = -size
        x = 0
        for i in range(size):
            d = self.r_short()
            x = x | (d << (i * 15))
        return x * sign

    dispatch[TYPE_LONG] = load_long

    def load_none(self) -> None:
        return None

    dispatch[TYPE_NONE] = load_none

    def load_set(self):
        n = self.r_long()
        args = [self.load() for i in range(n)]
        return set(args)

    dispatch[TYPE_SET] = load_set

    def load_stopiter(self) -> type[StopIteration]:
        return StopIteration

    dispatch[TYPE_STOPITER] = load_stopiter

    def load_string(self):
        n = self.r_long()
        return self._read(n)

    dispatch[TYPE_STRING] = load_string

    def load_stringref(self):
        n = self.r_long()
        return self._stringtable[n]

    dispatch[TYPE_STRINGREF] = load_stringref

    def load_true(self) -> bool:
        return True

    dispatch[TYPE_TRUE] = load_true

    def load_tuple(self):
        return tuple(self.load_list())

    dispatch[TYPE_TUPLE] = load_tuple

    def load_unicode(self):
        n = self.r_long()
        s = self._read(n)
        ret = s.decode("utf8")
        return ret

    dispatch[TYPE_UNICODE] = load_unicode


# ________________________________________________________________


def _read(self, n):
    pos = self.bufpos
    newpos = pos + n
    if newpos > len(self.bufstr):
        raise EOFError
    ret = self.bufstr[pos:newpos]
    self.bufpos = newpos
    return ret


def _read1(self):
    ret = self.bufstr[self.bufpos]
    self.bufpos += 1
    return ret


def _r_short(self):
    lo = _read1(self)
    hi = _read1(self)
    x = lo | (hi << 8)
    if x & 0x8000:
        x = x - 0x10000
    return x


def _r_long(self):
    # inlined this most common case
    p = self.bufpos
    s = self.bufstr
    a = s[p]
    b = s[p + 1]
    c = s[p + 2]
    d = s[p + 3]
    self.bufpos += 4
    x = a | (b << 8) | (c << 16) | (d << 24)
    if d & 0x80 and x > 0:
        x = -((1 << 32) - x)
        return int(x)
    else:
        return x


def _r_long64(self):
    a = _read1(self)
    b = _read1(self)
    c = _read1(self)
    d = _read1(self)
    e = _read1(self)
    f = _read1(self)
    g = _read1(self)
    h = _read1(self)
    x = a | (b << 8) | (c << 16) | (d << 24)
    x = x | (e << 32) | (f << 40) | (g << 48) | (h << 56)
    if h & 0x80 and x > 0:
        x = -((1 << 64) - x)
    return x


_load_dispatch = {}


class _FastUnmarshaller:
    dispatch = {}

    def __init__(self, buffer, python_version=None) -> None:
        self.bufstr = buffer
        self.bufpos = 0
        self._stringtable = []
        self.python_version = python_version

    def load(self):
        # make flow space happy
        c = "?"
        try:
            c = self.bufstr[self.bufpos]
            c = chr(c)
            self.bufpos += 1
            return _load_dispatch[c](self)
        except KeyError:
            exception = ValueError(
                "bad marshal code at position %d: %c" % (self.bufpos - 1, c)
            )
        except IndexError:
            exception = EOFError
        raise exception

    def load_null(self) -> type[_NULL]:
        return _NULL

    dispatch[TYPE_NULL] = load_null

    def load_none(self) -> None:
        return None

    dispatch[TYPE_NONE] = load_none

    def load_true(self) -> bool:
        return True

    dispatch[TYPE_TRUE] = load_true

    def load_false(self) -> bool:
        return False

    dispatch[TYPE_FALSE] = load_false

    def load_stopiter(self) -> type[StopIteration]:
        return StopIteration

    dispatch[TYPE_STOPITER] = load_stopiter

    def load_ellipsis(self) -> EllipsisType:
        return Ellipsis

    dispatch[TYPE_ELLIPSIS] = load_ellipsis

    def load_int(self):
        return _r_long(self)

    dispatch[TYPE_INT] = load_int

    def load_int64(self):
        return _r_long64(self)

    dispatch[TYPE_INT64] = load_int64

    def load_long(self):
        size = _r_long(self)
        sign = 1
        if size < 0:
            sign = -1
            size = -size
        x = 0
        for i in range(size):
            d = _r_short(self)
            x = x | (d << (i * 15))
        return x * sign

    dispatch[TYPE_LONG] = load_long

    def load_float(self) -> float:
        n = _read1(self)
        s = _read(self, n)
        return float(s)

    dispatch[TYPE_FLOAT] = load_float

    def load_complex(self) -> complex:
        n = _read1(self)
        s = _read(self, n)
        real = float(s)
        n = _read1(self)
        s = _read(self, n)
        imag = float(s)
        return complex(real, imag)

    dispatch[TYPE_COMPLEX] = load_complex

    def load_string(self):
        n = _r_long(self)
        return _read(self, n)

    dispatch[TYPE_STRING] = load_string

    def load_interned(self) -> str:
        n = _r_long(self)
        s = _read(self, n)
        s = s.decode("utf8")
        ret = intern(s)
        self._stringtable.append(ret)
        return ret

    dispatch[TYPE_INTERNED] = load_interned

    def load_stringref(self):
        n = _r_long(self)
        return self._stringtable[n]

    dispatch[TYPE_STRINGREF] = load_stringref

    def load_unicode(self):
        n = _r_long(self)
        s = _read(self, n)
        ret = s.decode("utf8")
        return ret

    dispatch[TYPE_UNICODE] = load_unicode

    def load_tuple(self):
        return tuple(self.load_list())

    dispatch[TYPE_TUPLE] = load_tuple

    def load_list(self):
        n = _r_long(self)
        list = []
        for i in range(n):
            list.append(self.load())
        return list

    dispatch[TYPE_LIST] = load_list

    def load_dict(self):
        d = {}
        while 1:
            key = self.load()
            if key is _NULL:
                break
            value = self.load()
            d[key] = value
        return d

    dispatch[TYPE_DICT] = load_dict

    def load_code(self) -> Code2 | CodeType:
        argcount = _r_long(self)
        nlocals = _r_long(self)
        stacksize = _r_long(self)
        flags = _r_long(self)
        code = self.load()
        consts = self.load()
        names = self.load()
        varnames = self.load()
        freevars = self.load()
        cellvars = self.load()
        filename = self.load()
        name = self.load()
        firstlineno = _r_long(self)
        lnotab = self.load()
        if isinstance(name, bytes):
            name = name.decode()
        return Code2(
            argcount,
            nlocals,
            stacksize,
            flags,
            code,
            consts,
            names,
            varnames,
            filename.decode(),
            name,
            firstlineno,
            lnotab,
            freevars,
            cellvars,
        )

    dispatch[TYPE_CODE] = load_code

    def load_set(self):
        n = _r_long(self)
        args = [self.load() for i in range(n)]
        return set(args)

    dispatch[TYPE_SET] = load_set

    def load_frozenset(self):
        n = _r_long(self)
        args = [self.load() for i in range(n)]
        return frozenset(args)

    dispatch[TYPE_FROZENSET] = load_frozenset


_load_dispatch = _FastUnmarshaller.dispatch

# _________________________________________________________________
#
# user interface

version = 1


@builtinify
def dump(
    x,
    f,
    version: int = version,
    python_version: tuple = PYTHON_VERSION_TRIPLE,
    is_pypy: Optional[bool] = None,
) -> None:
    # XXX 'version' is ignored, we always dump in a version-0-compatible format
    m = _Marshaller(f.write, python_version, is_pypy)
    m.dump(x)


@builtinify
def load(f, python_version: tuple = PYTHON_VERSION_TRIPLE, is_pypy=None):
    um = _Unmarshaller(f.read, python_version, is_pypy)
    return um.load()


@builtinify
def dumps(
    x,
    python_version: tuple[int, ...] = PYTHON_VERSION_TRIPLE,
    is_pypy: Optional[bool] = None,
) -> bytes | str:
    buffer = []
    collection_order = x.collection_order if hasattr(x, "collection_order") else {}
    reference_objects = (
        x.reference_objects if hasattr(x, "reference_objects") else set()
    )
    m = _Marshaller(
        buffer.append,
        python_version=python_version,
        is_pypy=is_pypy,
        collection_order=collection_order,
        reference_objects=reference_objects,
    )
    flag_ref = FLAG_REF if python_version >= (3, 4) else 0
    m.dump(x, flag_ref)
    if python_version:
        is_python3 = python_version >= (3, 0)
    else:
        is_python3 = True

    if is_python3:
        if PYTHON_VERSION_TRIPLE >= (3, 0):
            # Python 3.x handling  Python 3.x
            buf = []
            for b in buffer:
                if isinstance(b, str):
                    s2b = bytes(ord(b[j]) for j in range(len(b)))
                    buf.append(s2b)
                elif isinstance(b, bytearray):
                    buf.append(str(b))
                else:
                    buf.append(b)
            return b"".join(buf)
        else:
            # Python 2.x handling Python 3.x
            buf = b""
            for b in buffer:
                buf += b.decode(errors="ignore")
                pass
            return buf
    else:
        # Python 2 or 3 handling Python 2.x
        buf = []
        for b in buffer:
            if isinstance(b, str):
                if python_version < (2, 0):
                    # Python 1.x has no notion of Unicode. It uses strings.
                    buf.append(b)
                else:
                    try:
                        s2b = bytes(ord(b[j]) for j in range(len(b)))
                    except ValueError:
                        s2b = b.encode("utf-8")
                    buf.append(s2b)
            elif isinstance(b, bytearray):
                buf.append(str(b))
            else:
                buf.append(b)

        return b"".join(buf)


@builtinify
def loads(s, python_version=None):
    um = _FastUnmarshaller(s, python_version)
    return um.load()



# (C) Copyright 2025 by Rocky Bernstein
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
Common routines for entering and classifying opcodes. Inspired by,
limited by, and somewhat compatible with the corresponding
Python opcode.py structures
"""


def get_optype_graal(opcode: int, opc) -> str:
    """Helper to determine what class of instructions ``opcode`` is in.
    Return is a string in:
       compare, const, free, jabs, jrel, local, name, nargs, or ??
    """
    if opcode in opc.BINARY_OPS:
        return "binary"
    elif opcode in opc.COLLECTION_OPS:
        return "collection"
    elif opcode in opc.CONST_OPS:
        return "const"
    elif opcode in opc.COMPARE_OPS:
        return "compare"
    elif opcode in opc.ENCODED_ARG_OPS:
        return "encoded_arg"
    elif opcode in opc.FREE_OPS:
        return "free"
    elif opcode in opc.JABS_OPS:
        return "jabs"
    elif opcode in opc.JREL_OPS:
        return "jrel"
    # elif opcode in opc.LOCAL_OPS:
    #     return "local"
    elif opcode in opc.NAME_OPS:
        return "name"
    elif opcode in opc.NARGS_OPS:
        return "nargs"
    # This has to come after NARGS_OPS. Some are in both?
    elif opcode in opc.VARGS_OPS:
        return "vargs"
    elif opcode in opc.UNARY_OPS:
        return "unary"

    return "??"


def findlabels(bytecode, opc):
    """Returns a list of instruction offsets in the supplied bytecode
    which are the targets of some sort of jump instruction.
    """
    offset = 0
    n = len(bytecode)
    offsets = []
    while offset < n:
        opcode = bytecode[offset]
        optype = get_optype_graal(opcode, opc)

        # opname = opc.opname[opcode]
        # print(
        #     f"offset: {offset} {hex(opcode)} {opname} optype: {optype}"
        # )

        jump_offset = -1
        if optype == "jrel":
            arg = bytecode[offset + 1]
            if opcode == opc.opmap["JUMP_BACKWARD"]:
                jump_offset = offset - arg
            else:
                jump_offset = offset + arg
        # elif op in opc.JABS_OPS:
        #     jump_offset = arg
        if jump_offset >= 0:
            if jump_offset not in offsets:
                offsets.append(jump_offset)
        inst_size = opc.arg_counts[opcode] + 1
        offset += inst_size

    return offsets


# FIXME: Use an enumeration? Or tuple?
BINARY_OPS = {
    0: "ADD",
    1: "INPLACE_ADD",
    2: "SUB",
    3: "INPLACE_SUB",
    4: "MUL",
    5: "INPLACE_MUL",
    6: "FLOORDIV",
    7: "INPLACE_FLOORDIV",
    8: "TRUEDIV",
    9: "INPLACE_TRUEDIV",
    10: "MOD",
    11: "INPLACE_MOD",
    12: "EQ",
    13: "NE",
    14: "LT",
    15: "LE",
    16: "GT",
    17: "GE",
    18: "LSHIFT",
    19: "INPLACE_LSHIFT",
    20: "RSHIFT",
    21: "INPLACE_RSHIFT",
    22: "AND",
    23: "INPLACE_AND",
    24: "OR",
    25: "INPLACE_OR",
    26: "XOR",
    27: "INPLACE_XOR",
    28: "POW",
    29: "INPLACE_POW",
    30: "IN",
    31: "IS",
    32: "MATMUL",
    33: "INPLACE_MATMUL",
}

COLLECTION_KIND_SHIFT = 5
COLLECTION_KIND_MASK = (1 << COLLECTION_KIND_SHIFT) - 1
COLLECTION_KIND = {
    0b001: "list",
    0b010: "tuple",  # Probably need to use a mask
    0b011: "set",
    0b100: "dict",
    0b101: "PKeyword",
    0b110: "Object",
    # #
    # 99: "tuple",  # probably need to use a mask
    # 160: "object",
    # 192: "object",
}

COLLECTION_KIND_ELEMENT = {
    0: "",
    1: "int",
    2: "long",
    3: "boolean",
    4: "double",
    5: "object"
}


UNARY_OPS = {
    0: "NOT",
    31: "IS",
    30: "IN",
}

def collection_to_str(collection_value: int) -> str:
    kind = collection_value >> COLLECTION_KIND_SHIFT
    kind_str = COLLECTION_KIND.get(kind, "??")
    element_type = collection_value & COLLECTION_KIND_MASK
    element_str = COLLECTION_KIND_ELEMENT.get(element_type, "??")
    return f"{kind_str}[{element_str}]"

def binary_op_graal(
    loc: dict,
    name: str,
    opcode: int,
    pop: int = 2,
    push: int = 1,
    arg_count: int = 1,
) -> None:
    """
    Put opcode in the class of instructions that are binary operations.
    """
    loc["binaryop"].add(opcode)
    def_op_graal(loc, name, opcode, pop, push, arg_count)


def def_op_graal(
    loc: dict,
    op_name: str,
    opcode: int,
    pop: int = -2,
    push: int = -2,
    arg_count: int = 0,
    fallthrough: bool = True,
) -> None:
    loc["opname"][opcode] = op_name
    loc["opmap"][op_name] = opcode
    loc["oppush"][opcode] = push
    loc["oppop"][opcode] = pop
    if not fallthrough:
        loc["nofollow"].append(opcode)
    loc["arg_counts"][opcode] = arg_count
    loc["operator_set"] = frozenset(loc["nullaryop"] | loc["unaryop"] | loc["binaryop"])


def call_op_graal(
    loc: dict,
    name: str,
    opcode: int,
    pop: int = -2,
    push: int = 1,
    fallthrough: bool = True,
) -> None:
    """
    Put opcode in the class of instructions that perform calls.
    """
    loc["callop"].add(opcode)
    nargs_op_graal(loc, name, opcode, pop, push, fallthrough)


def collection_op_graal(
    loc: dict, name: str, opcode: int, pop: int = 0, push: int = 1, arg_count: int = 1
) -> None:
    def_op_graal(loc, name, opcode, pop, push, arg_count)
    loc["collectionop"].add(opcode)


def const_op_graal(
    loc: dict, name: str, opcode: int, pop: int = 0, push: int = 1, arg_count: int = 1
) -> None:
    def_op_graal(loc, name, opcode, pop, push, arg_count)
    loc["hasconst"].append(opcode)
    loc["nullaryop"].add(opcode)


def free_op_graal(
    loc: dict, name: str, opcode: int, pop: int = 0, push: int = 1, arg_count: int = 1
) -> None:
    def_op_graal(loc, name, opcode, pop, push, arg_count)
    loc["hasfree"].append(opcode)


def jrel_op_graal(
    loc,
    name: str,
    opcode: int,
    pop: int = 0,
    push: int = 0,
    conditional=False,
    fallthrough=True,
    arg_count: int = 1,
) -> None:
    """
    Put opcode in the class of instructions that can perform a relative jump.
    """
    def_op_graal(loc, name, opcode, pop, push, arg_count)
    loc["hasjrel"].append(opcode)
    if conditional:
        loc["hascondition"].append(opcode)


def name_op_graal(
    loc: dict, op_name, opcode: int, pop=-2, push=-2, arg_count: int = 1
) -> None:
    """
    Put opcode in the class of instructions that index into the "name" table.
    """
    def_op_graal(loc, op_name, opcode, pop, push, arg_count)
    loc["hasname"].append(opcode)
    loc["nullaryop"].add(opcode)


def nargs_op_graal(
    loc,
    name: str,
    opcode: int,
    pop: int = -2,
    push: int = -1,
    arg_count: int = 0,
    fallthrough=True,
) -> None:
    """
    Put opcode in the class of instructions that have a variable number of (or *n*) arguments
    """
    def_op_graal(
        loc, name, opcode, pop, push, arg_count=arg_count, fallthrough=fallthrough
    )
    loc["hasnargs"].append(opcode)


def store_op_graal(
    loc: dict, name: str, op, pop=0, push=1, is_type="def", arg_count: int = 1
) -> None:
    if is_type == "name":
        name_op_graal(loc, name, op, pop, push, arg_count)
        loc["nullaryop"].remove(op)
    # elif is_type == "local":
    #     local_op(loc, name, op, pop, push)
    #     loc["nullaryop"].remove(op)
    # elif is_type == "free":
    #     free_op(loc, name, op, pop, push)
    else:
        assert is_type == "def"
        def_op_graal(loc, name, op, pop, push)
    loc["hasstore"].append(op)


def unary_op_graal(
    loc: dict,
    name: str,
    opcode: int,
    pop: int = 1,
    push: int = 1,
    arg_count: int = 1,
) -> None:
    """
    Put opcode in the class of instructions that are binary operations.
    """
    loc["unaryop"].add(opcode)
    def_op_graal(loc, name, opcode, pop, push, arg_count)


def update_sets(loc) -> None:
    """
    Updates various category sets all opcode have been defined.
    """

    loc["BINARY_OPS"] = frozenset(loc["binaryop"])
    loc["COLLECTION_OPS"] = frozenset(loc["collectionop"])
    loc["COMPARE_OPS"] = frozenset(loc["hascompare"])
    loc["CONDITION_OPS"] = frozenset(loc["hascondition"])
    loc["CONST_OPS"] = frozenset(loc["hasconst"])
    loc["EXTENDED_ARG"] = loc["opmap"]["EXTENDED_ARG"]
    loc["ENCODED_ARG_OPS"] = frozenset(loc["encoded_arg"])
    loc["FREE_OPS"] = frozenset(loc["hasfree"])
    loc["JREL_OPS"] = frozenset(loc["hasjrel"])
    loc["JABS_OPS"] = frozenset(loc["hasjabs"])
    loc["NAME_OPS"] = frozenset(loc["hasname"])
    loc["NARGS_OPS"] = frozenset(loc["hasnargs"])
    loc["VARGS_OPS"] = frozenset(loc["hasvargs"])
    loc["UNARY_OPS"] = frozenset(loc["unaryop"])
    loc["findlabels"] = findlabels
    loc["get_jump_targets"] = findlabels



# (C) Copyright 2017, 2020-2021, 2023, 2025 by Rocky Bernstein
"""
CPython 3.3 bytecode opcodes

This is a like Python 3.3's opcode.py with some classification
of stack usage.
"""

from xdis.opcodes.base import (
    VARYING_STACK_INT,
    binary_op,
    call_op,
    compare_op,
    const_op,
    cpython_implementation,
    def_op,
    finalize_opcodes,
    free_op,
    init_opdata,
    jabs_op,
    jrel_op,
    local_op,
    name_op,
    nargs_op,
    rm_op,
    store_op,
    unary_op,
    update_pj3,
    varargs_op,
)
from xdis.opcodes.format.basic import (
    format_extended_arg,
    format_RAISE_VARARGS_older,
    opcode_arg_fmt_base,
)
from xdis.opcodes.format.extended import opcode_extended_fmt_base
from xdis.opcodes.opcode_3x.opcode_3x import format_MAKE_FUNCTION_30_35

version_tuple = (3, 3)
python_implementation = cpython_implementation

loc = locals()
init_opdata(loc, None, version_tuple)

# Below are opcode changes since Python 3.2

# fmt: off
# Instruction opcodes for compiled code
# Blank lines correspond to available opcodes

# If the POP field is -1 and the opcode is a var args operation
# then the operand holds the size.
#
# If the POP field is negative and the opcode is a nargs operation
# then pop the operand amount plus the negative of the POP amount.

# fmt: off
#              OP NAME               OPCODE POP PUSH
#--------------------------------------------------
def_op(loc,   "POP_TOP",                  1,  1,  0)
def_op(loc,   "ROT_TWO",                  2,  2,  2)
def_op(loc,   "ROT_THREE",                3,  3,  3)
def_op(loc,   "DUP_TOP",                  4,  0,  1)

# Python 3.2+
def_op(loc,   "DUP_TOP_TWO",              5,  0,  2)

def_op(loc,   "NOP",                      9,  0,  0)
unary_op(loc, "UNARY_POSITIVE",          10)
unary_op(loc, "UNARY_NEGATIVE",          11)
unary_op(loc, "UNARY_NOT",               12)

unary_op(loc, "UNARY_INVERT",            15)

binary_op(loc, "BINARY_POWER",           19)
binary_op(loc, "BINARY_MULTIPLY",        20)

binary_op(loc, "BINARY_MODULO",          22)
binary_op(loc, "BINARY_ADD",             23)
binary_op(loc, "BINARY_SUBTRACT",        24)
binary_op(loc, "BINARY_SUBSCR",          25)
binary_op(loc, "BINARY_FLOOR_DIVIDE",    26)
binary_op(loc, "BINARY_TRUE_DIVIDE",     27)
binary_op(loc, "INPLACE_FLOOR_DIVIDE",   28)
binary_op(loc, "INPLACE_TRUE_DIVIDE",    29)

# Gone from Python 3 are Python2's
# SLICE+0 ... SLICE+3
# STORE_SLICE+0 ... STORE_SLICE+3
# DELETE_SLICE+0 ... DELETE_SLICE+3

#          OP NAME                OPCODE POP PUSH
#-----------------------------------------------
store_op(loc, "STORE_MAP",            54,  3,  1)
binary_op(loc, "INPLACE_ADD",            55,  2,  1)
binary_op(loc, "INPLACE_SUBTRACT",       56,  2,  1)
binary_op(loc, "INPLACE_MULTIPLY",       57,  2,  1)

binary_op(loc, "INPLACE_MODULO",         59,  2,  1)
store_op(loc, "STORE_SUBSCR",         60,  3,  0) # Implements TOS1[TOS] = TOS2.
def_op(loc, "DELETE_SUBSCR",          61,  2,  0) # Implements del TOS1[TOS].
binary_op(loc, "BINARY_LSHIFT",       62)
binary_op(loc, "BINARY_RSHIFT",       63)
binary_op(loc, "BINARY_AND",          64)
binary_op(loc, "BINARY_XOR",          65)
binary_op(loc, "BINARY_OR",           66)
binary_op(loc, "INPLACE_POWER",       67)
def_op(loc, "GET_ITER",               68,  1,  1)
store_op(loc, "STORE_LOCALS",         69,  1,  0)

def_op(loc, "PRINT_EXPR",             70,  1,  0)
unary_op(loc, "LOAD_BUILD_CLASS",     71,  0,  1)

# Python3 drops/changes:
#  def_op(loc, "PRINT_ITEM", 71)
#  def_op(loc, "PRINT_NEWLINE", 72)
#  def_op(loc, "PRINT_ITEM_TO", 73)
#  def_op(loc, "PRINT_NEWLINE_TO", 74)

binary_op(loc, "INPLACE_LSHIFT",      75)
binary_op(loc, "INPLACE_RSHIFT",      76)
binary_op(loc, "INPLACE_AND",         77)
binary_op(loc, "INPLACE_XOR",         78)
binary_op(loc, "INPLACE_OR",          79)
def_op(loc, "BREAK_LOOP",             80,  0,  0, fallthrough=False)
def_op(loc, "WITH_CLEANUP",           81,  1,  0) # Cleans up the stack when a with statement
                                                  # block exits.  Handle stack special

def_op(loc, "RETURN_VALUE",           83,  1,  0, fallthrough=False)
def_op(loc, "IMPORT_STAR",            84,  1,  0)

def_op(loc, "YIELD_VALUE",            86,  1,  1)
def_op(loc, "POP_BLOCK",              87,  0,  0)
def_op(loc, "END_FINALLY",            88,  1,  0)
def_op(loc, "POP_EXCEPT",             89,  0,  0)

HAVE_ARGUMENT = 90              # Opcodes from here have an argument:

#          OP NAME                OPCODE POP PUSH
#-----------------------------------------------
store_op(loc, "STORE_NAME",           90,  1,  0, is_type="name")   # Operand is in name list
name_op(loc, "DELETE_NAME",           91,  0,  0)   # ""
varargs_op(loc, "UNPACK_SEQUENCE",    92,  1, VARYING_STACK_INT)  # unpacks TOS, arg is the count
jrel_op(loc,    "FOR_ITER",           93,  0,  1)

varargs_op(loc, "UNPACK_EX",          94,  0,  0)  # assignment with a starred target; arg is count
store_op(loc, "STORE_ATTR",           95,  2,  0, is_type="name")   # Operand is in name list
name_op(loc, "DELETE_ATTR",           96,  1,  0)   # ""
store_op(loc, "STORE_GLOBAL",         97,  1,  0, is_type="name")   # ""
name_op(loc, "DELETE_GLOBAL",         98,  0,  0)   # ""

# Python 2"s DUP_TOPX is gone starting in Python 3.2

#          OP NAME                OPCODE POP PUSH
#-----------------------------------------------
const_op(loc,   "LOAD_CONST",        100,  0,  1)  # Operand is in const list
loc["nullaryloadop"].add(100)

name_op(loc,    "LOAD_NAME",         101,  0,  1)  # Operand is in name list
loc["nullaryloadop"].add(101)

varargs_op(loc, "BUILD_TUPLE",       102, -1,  1)  # TOS is count of tuple items
varargs_op(loc, "BUILD_LIST",        103, -1,  1)  # TOS is count of list items
varargs_op(loc, "BUILD_SET",         104, -1,  1)  # TOS is count of set items
varargs_op(loc, "BUILD_MAP",         105,  0,  1)  # argument is dictionary count to be pushed
name_op(loc, "LOAD_ATTR",            106,  1,  1)  # Operand is in name list
compare_op(loc, "COMPARE_OP",        107,  2,  1)  # Comparison operator
name_op(loc, "IMPORT_NAME",          108,  2,  1)  # Imports TOS and TOS1; module pushed
loc["nullaryloadop"].add(108)

name_op(loc, "IMPORT_FROM",          109,  0,  1)  # Operand is in name list

jrel_op(loc, "JUMP_FORWARD",         110,  0,  0, fallthrough=False)  # Number of bytes to skip
jabs_op(loc, "JUMP_IF_FALSE_OR_POP", 111, conditional=True)  # Target byte offset from beginning of code
jabs_op(loc, "JUMP_IF_TRUE_OR_POP",  112, conditional=True) # ""
jabs_op(loc, "JUMP_ABSOLUTE",        113,  0,  0, fallthrough=False)  # Target byte offset from beginning of code
jabs_op(loc, "POP_JUMP_IF_FALSE",    114,  2,  1, conditional=True) # ""
jabs_op(loc, "POP_JUMP_IF_TRUE",     115,  2,  1, conditional=True) # ""

name_op(loc, "LOAD_GLOBAL",          116,  0,  1)  # Operand is in name list
loc["nullaryloadop"].add(116)


#          OP NAME                OPCODE POP PUSH
#-----------------------------------------------
jabs_op(loc, "CONTINUE_LOOP",        119,  0,  0, fallthrough=False)  # Target address
jrel_op(loc, "SETUP_LOOP",           120,  0,  0, conditional=True) # Distance to target address
jrel_op(loc, "SETUP_EXCEPT",         121,  0,  6, conditional=True)  # ""
jrel_op(loc, "SETUP_FINALLY",        122,  0,  6, conditional=True)  # ""

local_op(loc, "LOAD_FAST",           124,  0,  1)  # Local variable number
loc["nullaryloadop"].add(124)

store_op(loc, "STORE_FAST",          125,  1,  0, is_type="local")  # Local variable number
local_op(loc, "DELETE_FAST",         126,  0,  0)  # Local variable number

nargs_op(loc, "RAISE_VARARGS",       130, -1,  1, fallthrough=False)
                                                 # Number of raise arguments (1, 2, or 3)

call_op(loc, "CALL_FUNCTION",        131, -1,  1)  # #args + (#kwargs << 8)

nargs_op(loc, "MAKE_FUNCTION",       132, -2,  1) # TOS is number of args if < 3.6
varargs_op(loc, "BUILD_SLICE",       133,  2,  1) # TOS is number of items to pop

nargs_op(loc, "MAKE_CLOSURE",        134, -3,  1) # TOS is number of items to pop
free_op(loc, "LOAD_CLOSURE",         135,  0,  1)
loc["nullaryloadop"].add(135)

loc["nullaryloadop"].add(136)


free_op(loc, "LOAD_DEREF",           136,  0,  1)
loc["nullaryloadop"].add(136)

loc["nullaryop"].add(136)
loc["nullaryloadop"].add(136)

store_op(loc, "STORE_DEREF",         137,  1,  0, is_type="free")
free_op(loc, "DELETE_DEREF",         138,  0,  0)

#          OP NAME                OPCODE POP PUSH
#-----------------------------------------------
nargs_op(loc, "CALL_FUNCTION_VAR",   140,  -2,  1)  # #args + (#kwargs << 8)
nargs_op(loc, "CALL_FUNCTION_KW",    141,  -2,  1)  # #args + (#kwargs << 8)
nargs_op(loc, "CALL_FUNCTION_VAR_KW",142,  -3,  1)   # #args + (#kwargs << 8)

jrel_op(loc, "SETUP_WITH",           143,  0,  7)

def_op(loc, "LIST_APPEND",           145,  2,  1)  # Calls list.append(TOS[-i], TOS).
                                                 # Used to implement list comprehensions.
def_op(loc, "SET_ADD",               146,  1,  0)  # Calls set.add(TOS1[-i], TOS).
                                                 # Used to implement set comprehensions.
def_op(loc, "MAP_ADD",               147,  3,  1)  # Calls dict.setitem(TOS1[-i], TOS, TOS1)
                                                 # Used to implement dict comprehensions.

def_op(loc, "EXTENDED_ARG",          144,  0,   0)
# fmt: on

EXTENDED_ARG = 144

opcode_arg_fmt = {"EXTENDED_ARG": format_extended_arg}

def_op(loc, "YIELD_FROM", 72, 1, 0)
# fmt: on


opcode_extended_fmt = opcode_extended_fmt33 = opcode_extended_fmt_base.copy()
opcode_arg_fmt = opcode_arg_fmt33 = {
    **opcode_arg_fmt_base,
    **{
        "MAKE_CLOSURE": format_MAKE_FUNCTION_30_35,
        "MAKE_FUNCTION": format_MAKE_FUNCTION_30_35,
        "RAISE_VARARGS": format_RAISE_VARARGS_older,
    },
}

update_pj3(globals(), loc)
finalize_opcodes(loc)



# (C) Copyright 2017, 2020, 2023-2025 by Rocky Bernstein
"""
PYPY 3.5 opcodes

This is a like Python 3.5's opcode.py with some classification
of stack usage.
"""

import xdis.opcodes.opcode_3x.opcode_35 as opcode_35
from xdis.opcodes.base import (
    call_op,
    def_op,
    finalize_opcodes,
    init_opdata,
    jrel_op,
    name_op,
    update_pj3,
    varargs_op,
)
from xdis.opcodes.format.extended import (
    extended_format_ATTR,
    extended_format_CALL_METHOD,
    extended_format_RETURN_VALUE,
)
from xdis.opcodes.opcode_3x.opcode_36 import extended_format_BUILD_STRING
from xdis.version_info import PythonImplementation

version_tuple = (3, 5)
python_implementation = PythonImplementation("PyPy")

loc = locals()
init_opdata(loc, opcode_35, version_tuple, is_pypy=True)

## FIXME: DRY common PYPY opcode additions

# PyPy only
# ----------
def_op(loc, "FORMAT_VALUE", 155)
varargs_op(loc, "BUILD_STRING", 157)
name_op(loc, "LOOKUP_METHOD", 201, 1, 2)
call_op(loc, "CALL_METHOD", 202, -1, 1)
loc["hasvargs"].append(202)

# Used only in single-mode compilation list-comprehension generators
varargs_op(loc, "BUILD_LIST_FROM_ARG", 203)

# Used only in assert statements
jrel_op(loc, "JUMP_IF_NOT_DEBUG", 204, conditional=True)

# Python 3.5.3 PyPYy 7.0.0 adds LOAD_REVDB.  We can't distinguish
# between the two kinds of bytecode by Python version number, or magic
# number. And we don't have a means of specifying
# platform.python_branch() which would give us the 7.0.0 as opposed to
# 6.0.0.

# Lacking anything better to do, we'll add this opcode since
# the newest PyPy for 3.5 has it.

def_op(loc, "LOAD_REVDB_VAR", 205)

opcode_extended_fmt = {
    "BUILD_STRING": extended_format_BUILD_STRING,
    "CALL_METHOD": extended_format_CALL_METHOD,
    "LOAD_ATTR": extended_format_ATTR,
    "RETURN_VALUE": extended_format_RETURN_VALUE,
    "STORE_ATTR": extended_format_ATTR,
}

# FIXME remove (fix uncompyle6)
update_pj3(globals(), loc)
finalize_opcodes(loc)



#!/usr/bin/env python
# (C) Copyright 2025 by Rocky Bernstein
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

Check that we can read in or unmarshal a Python bytecode file and
then write it or marshal it back and the files are identical.

As
Input is a Python bytecode file. A file of this kind usually the has
file extension ".pyc" or ".pyo".

This steps taken are:
- loads it or unmarshals it with xdis.load.load_module_from_file_object()
- writes it or marshals back out using xdis.load.write_bytecode_file() to a temporary file
- compares the original and written files by raw bytes and by the
  module/code-object structure they contain

Usage:
   test_writing_pyc.py path/to/bytecode_file.pyc

"""

import filecmp
import os
import os.path as osp
import sys
import tempfile
import types
from typing import Union

from xdis.codetype.base import CodeBase
from xdis.load import load_module_from_file_object, write_bytecode_file
from xdis.version_info import version_tuple_to_str

# For later:
# from xdis.unmarshal import (
#     _VersionIndependentUnmarshaller,
#     FLAG_REF,
#     UNMARSHAL_DISPATCH_TABLE,
# )


def compare_consts(c1: tuple, c2: tuple) -> bool:
    """Compare tuples of constants, recursing into code objects when found."""
    if len(c1) != len(c2):
        return False
    for a, b in zip(c1, c2):
        # If code object-like, compare as code objects
        if hasattr(a, "co_code") and hasattr(b, "co_code"):
            if not compare_code_objects(a, b):
                return False
        else:
            if a != b:
                return False
    return True

def compare_showing_error(orig_path: str, new_path: str):
    """
      Compare file contents sizes if those mismatch and the first hex offset that the differ.
    """
    orig_bytes = open(orig_path, "rb").read()
    new_bytes = open(new_path, "rb").read()
    if (orig_n := len(orig_bytes)) != (new_n := len(new_bytes)):
        print(f"MISMATCH: original has {orig_n} bytes; new has {new_n} bytes", file=sys.stderr)

    for i, (old_byte, new_byte) in enumerate(zip(orig_bytes, new_bytes)):
        if (old_byte != new_byte):
            print(f"MISMATCH at {hex(i)}: old {hex(old_byte)}; new: {hex(new_byte)}", file=sys.stderr)
            return


def get_code_attrs(co: Union[CodeBase, types.CodeType]) -> dict:
    """
    Extract a set of attributes from a code object that are stable for comparison.
    This is defensive: only include an attribute if it exists on the object.
    """
    attrs = {}
    for name in (
        "co_argcount",
        "co_posonlyargcount",
        "co_kwonlyargcount",
        "co_nlocals",
        "co_stacksize",
        "co_flags",
        "co_code",
        "co_consts",
        "co_names",
        "co_varnames",
        "co_freevars",
        "co_cellvars",
        "co_filename",
        "co_name",
    ):
        if hasattr(co, name):
            attrs[name] = getattr(co, name)
    return attrs


def compare_code_objects(
    a: Union[CodeBase, types.CodeType], b: Union[CodeBase, types.CodeType]
) -> bool:
    """
    Compare two code objects by a selection of attributes.
    This function attempts to be permissive across Python implementations
    by only using attributes when present and by comparing constants recursively.
    """
    attrs_a = get_code_attrs(a)
    attrs_b = get_code_attrs(b)

    if set(attrs_a.keys()) != set(attrs_b.keys()):
        # If the sets of attributes differ, still try to compare the
        # intersection of attributes to be less strict across implementations.
        common_keys = sorted(set(attrs_a.keys()).intersection(attrs_b.keys()))
    else:
        common_keys = sorted(attrs_a.keys())

    for key in common_keys:
        va = attrs_a[key]
        vb = attrs_b[key]
        if key == "co_consts":
            if not compare_consts(tuple(va), tuple(vb)):
                return False
        else:
            if va != vb:
                return False
    return True


# For later
# def r_object(self, bytes_for_s: bool = False):
#     """
#     Replacement r_object for classification
#     """
#     byte1 = ord(self.fp.read(1))

#     # FLAG_REF indicates whether we "intern" or
#     # save a reference to the object.
#     # byte1 without that reference is the
#     # marshal type code, an ASCII character.
#     save_ref = False
#     if byte1 & FLAG_REF:
#         # Since 3.4, "flag" is the marshal.c name
#         save_ref = True
#         byte1 = byte1 & (FLAG_REF - 1)
#     marshal_type = chr(byte1)

#     # print(marshal_type)  # debug

#     if marshal_type in UNMARSHAL_DISPATCH_TABLE:
#         func_suffix = UNMARSHAL_DISPATCH_TABLE[marshal_type]
#         unmarshal_func = getattr(self, "t_" + func_suffix)
#         return unmarshal_func(save_ref, bytes_for_s)
#     else:
#         mess = ("Unknown type %i (hex %x) %c\n"
#                 % (ord(marshal_type), ord(marshal_type), marshal_type))
#         raise TypeError(mess)
#     return

def load_meta_and_code_from_filename(path: str):
    """
    Open path and use load_module_from_file_object to get:
    (version_tuple, timestamp, magic_int, co, is_pypy, source_size, sip_hash, file_offsets)
    """
    with open(path, "rb") as fp:
        return load_module_from_file_object(fp, filename=path, get_code=True)


def roundtrip_pyc(input_path: str, unlink_on_success: bool, verbose: bool) -> int:

    # parser = argparse.ArgumentParser(
    #     description="Load a .pyc with xdis, rewrite it to a temporary file, and compare."
    # )
    # parser.add_argument("pycfile", help="Path to the .pyc (or other bytecode) file")
    # args = parser.parse_args(argv)

    if not osp.exists(input_path):
        print(f"ERROR: file does not exist: {input_path}", file=sys.stderr)
        return 2
    if not osp.isfile(input_path):
        print(f"ERROR: not a file: {input_path}", file=sys.stderr)
        return 2

    # Load original using the file-object loader (it will close the file for us)
    try:
        (
            orig_version,
            orig_timestamp,
            orig_magic_int,
            orig_co,
            orig_is_pypy,
            orig_source_size,
            orig_sip_hash,
            orig_file_offsets,
        ) = load_meta_and_code_from_filename(input_path)
    except Exception as e:
        print(f"ERROR: failed to load original bytecode file '{input_path}:\n\t{e}", file=sys.stderr)
        return 3

    tf_name_base = osp.basename(input_path)
    if tf_name_base.endswith(".pyc"):
        tf_name_base = tf_name_base[: -len(".pyc")]
    elif tf_name_base.endswith(".pyc"):
        tf_name_base = tf_name_base[: -len(".pyo")]

    version_str = version_tuple_to_str(orig_version)
    tf_name_base += f"-{version_str}-"

    # Write to a temporary file using write_bytecode_file
    tf = tempfile.NamedTemporaryFile(prefix=tf_name_base, suffix=".pyc", delete=False)
    tf_name = tf.name
    tf.close()  # write_bytecode_file will open/write the file itself

    try:
        write_bytecode_file(
            tf_name,
            orig_co,
            orig_magic_int,
            compilation_ts=orig_timestamp,
            filesize=orig_source_size or 0,
            allow_native=False,
        )
    except TypeError:
        # Older/newer signatures might name the timestamp param differently; try without names
        try:
            write_bytecode_file(
                tf_name, orig_co, orig_magic_int, orig_timestamp, orig_source_size or 0
            )
        except Exception as e:
            print(f"ERROR: failed to write bytecode file '{tf_name}':\n\t{e}", file=sys.stderr)
            # Cleanup
            try:
                os.unlink(tf_name)
            except Exception:
                pass
            return 4
    except Exception as e:
        print(f"ERROR: failed to write bytecode file '{tf_name}:\n\t{e}", file=sys.stderr)
        try:
            os.unlink(tf_name)
        except Exception:
            pass
        return 4

    # Compare raw bytes first
    same_bytes = False
    try:
        same_bytes = filecmp.cmp(input_path, tf_name, shallow=False)
    except Exception as e:
        print(f"WARNING: could not do raw byte comparison: {e}", file=sys.stderr)

    print("Input file:", input_path)
    if verbose:
        print("Rewritten file:", tf_name)
        print("Raw-bytes identical:", same_bytes)
    if same_bytes:
        if unlink_on_success:
            os.unlink(tf_name)
        return 0

    compare_showing_error(input_path, tf_name)

    # Now compare by loading both and comparing metadata and code-object structure
    try:
        (
            new_version,
            new_timestamp,
            new_magic_int,
            new_co,
            new_is_pypy,
            new_source_size,
            new_sip_hash,
            new_file_offsets,
        ) = load_meta_and_code_from_filename(tf_name)
    except Exception as e:
        print(
            f"ERROR: failed to load rewritten bytecode file {tf_name}:\n\t{e}",
            file=sys.stderr,
        )
        return 5

    meta_equal = (
        orig_version == new_version
        and orig_magic_int == new_magic_int
        and (orig_timestamp == new_timestamp)
        and (orig_source_size == new_source_size)
        and (orig_sip_hash == new_sip_hash)
    )

    print(
        "Metadata equal (version, magic, timestamp, source_size, sip_hash):", meta_equal
    )
    print("Original is PyPy:", orig_is_pypy, "Rewritten is PyPy:", new_is_pypy)

    # Compare code objects
    codes_equal = compare_code_objects(orig_co, new_co)
    print("Code objects structurally equal:", codes_equal)

    # Compare file_offsets if present
    offsets_equal = orig_file_offsets == new_file_offsets
    print("File offsets equal:", offsets_equal)

    # all_identical = same_bytes and meta_equal and codes_equal and offsets_equal
    # if all_identical:
    #     print(
    #         "RESULT: files are identical both as raw bytes and in contained module/code objects."
    #     )
    #     ret = 0
    # else:
    #     print("RESULT: files differ.")
    #     # Helpful diagnostic: show which pieces disagree
    #     if not same_bytes:
    #         print("- raw-bytes differ")
    #     if not meta_equal:
    #         print("- metadata differs")
    #         print(
    #             "  original:",
    #             (
    #                 orig_version,
    #                 orig_magic_int,
    #                 orig_timestamp,
    #                 orig_source_size,
    #                 orig_sip_hash,
    #             ),
    #         )
    #         print(
    #             "  rewritten:",
    #             (
    #                 new_version,
    #                 new_magic_int,
    #                 new_timestamp,
    #                 new_source_size,
    #                 new_sip_hash,
    #             ),
    #         )
    #     if not codes_equal:
    #         print("- code objects differ (some attributes compared):")
    #         # attempt to print a short diff of co_code lengths and names
    #         try:
    #             print("  orig co_code len:", len(getattr(orig_co, "co_code", b"")))
    #             print("  new  co_code len:", len(getattr(new_co, "co_code", b"")))
    #             print("  orig co_names:", getattr(orig_co, "co_names", None))
    #             print("  new  co_names:", getattr(new_co, "co_names", None))
    #         except Exception:
    #             pass
    #     if not offsets_equal:
    #         print("- file offsets differ")
    #     ret = 1

    # Do not delete the temporary file automatically; leave it for inspection.
    print("Temporary rewritten file left at:", tf_name)
    return 1

def main() -> int:
    bad_count = 0
    for file in sys.argv[1:]:
        bad_count += roundtrip_pyc(file, unlink_on_success=False)
    return bad_count


if __name__ == "__main__":
    main()



# (C) Copyright 2017-2018, 2020-2021, 2023-2025
# by Rocky Bernstein
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""
CPython 3.2 bytecode opcodes to be used as a base for other opcodes including 3.2.

This is used in bytecode disassembly among other things. This is
similar to the opcodes in Python's opcode.py library.

If this file changes the other opcode files may have to be adjusted accordingly.
"""

from typing import Optional, Tuple

from xdis.opcodes.format.extended import opcode_extended_fmt_base, short_code_repr


def extended_format_MAKE_FUNCTION_30_35(opc, instructions) -> Tuple[Optional[str], int]:
    """make_function_inst should be a "MAKE_FUNCTION" or "MAKE_CLOSURE" instruction. TOS
    should have the function or closure name.
    """
    # From opcode description: argc indicates the total number of
    # positional and keyword arguments.  Sometimes the function name
    # is in the stack arg positions back.
    assert len(instructions) >= 2
    inst = instructions[0]
    assert inst.opname in ("MAKE_FUNCTION", "MAKE_CLOSURE")
    s = ""
    name_inst = instructions[1]
    start_offset = name_inst.offset
    if name_inst.opname in ("LOAD_CONST",):
        s += f"make_function({short_code_repr(name_inst.argval)}"
        return s, start_offset
    s += format_MAKE_FUNCTION_30_35(inst.argval)
    return s, start_offset


def format_MAKE_FUNCTION_30_35(argc: int) -> str:
    pos_args, name_pair_args, annotate_args = parse_fn_counts_30_35(argc)
    if (pos_args, name_pair_args, annotate_args) == (0, 0, 0):
        return "No arguments"

    s = "%d positional, %d keyword only, %d annotated" % (
        pos_args,
        name_pair_args,
        annotate_args,
    )
    return s


def parse_fn_counts_30_35(argc: int) -> Tuple[int, int, int]:
    """
    In Python 3.0 to 3.5 MAKE_CLOSURE and MAKE_FUNCTION encode
    arguments counts of positional, default + named, and annotation
    arguments a particular kind of encoding where each of
    the entry is a packed byte value of the lower 24 bits
    of ``argc``.  The high bits of argc may have come from
    an EXTENDED_ARG instruction. Here, we unpack the values
    from the ``argc`` int and return a triple of the
    positional args, named_args, and annotation args.
    """
    annotate_count = (argc >> 16) & 0x7FFF
    # For some reason that I don't understand, annotate_args is off by one
    # when there is an EXENDED_ARG instruction from what is documented in
    # https://docs.python.org/3.4/library/dis.html#opcode-MAKE_CLOSURE
    if annotate_count > 1:
        annotate_count -= 1
    return ((argc & 0xFF), (argc >> 8) & 0xFF, annotate_count)


opcode_extended_fmt_base3x = opcode_extended_fmt_base.copy()
